import numpy as np
import pickle as pk
from datetime import datetime
import pandas as pd
from fair_dynamic_rec.core.util.utils import get_dict_key_from_value
from tqdm import tqdm
import sys

class DataCmd:
    """
    Load the data files
    """

    def __init__(self, config, log_filename=None):
        # key: originalid, value: mappedid
        self.userid_mapped_data = {}
        self.itemid_mapped_data = {}
        self.supplierid_mapped_data = {}

        self.user_attribute_mapped_data = {}
        self.item_topic_mapped_data = {}

        # key: category (e.g. genre), value: integer index
        self.topic_mapped_data = {}

        self.rating_file_path = config.get_rating_file_path()
        self.train_file_path = config.get_train_file_path()
        self.test_file_path = config.get_test_file_path()
        self.user_file_path = config.get_user_file_path()
        self.item_file_path = config.get_item_file_path()
        self.supplier_file_path = config.get_supplier_file_path()

        print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Start loading data.")
        self.load_data_from_text_file(config)
        print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": data loaded.")

        self.n_users = self.ratings.shape[0] if self.ratings is not None else self.train_data.shape[0]
        self.n_items = self.ratings.shape[1] if self.ratings is not None else self.train_data.shape[1]

        self.feature_data = self.load_features(config)
        self.save_item_features(config)


    # def read_rating_as_pandas_dataframe(self, file_path, separator):
    #     return pd.read_csv(file_path, sep=separator, names=['userid', 'itemid', 'rating'], header=None)

    def read_rating_text_file(self, file_path, delimiter):
        data = np.array(np.zeros((1, 1)))
        user_index, item_index = 0, 0
        with open(file_path) as f:
            for instance in tqdm(f):
                instance_separated = instance.strip().split(delimiter)
                userid = instance_separated[0]
                itemid = instance_separated[1]
                rating = instance_separated[2]

                # map userid and itemid, and update the dictionary
                if userid not in self.userid_mapped_data.keys():
                    self.userid_mapped_data[userid] = user_index
                    user_index += 1
                if itemid not in self.itemid_mapped_data.keys():
                    self.itemid_mapped_data[itemid] = item_index
                    item_index += 1

                # update the data by appending the new data instance to the data matrix
                if (data.shape[0] - 1) < self.userid_mapped_data[userid]:
                    data = np.append(data, np.zeros((1, data.shape[1])), axis=0)
                if (data.shape[1] - 1) < self.itemid_mapped_data[itemid]:
                    data = np.append(data, np.zeros((1, data.shape[0])).transpose(), axis=1)

                data[self.userid_mapped_data[userid], self.itemid_mapped_data[itemid]] = rating

        return data
    def read_rating_text_file_given_size(self, file_path, delimiter, n_rows, n_cols):
        data = np.array(np.zeros((n_rows, n_cols)))
        user_index, item_index = 0, 0
        with open(file_path) as f:
            for instance in tqdm(f):
                instance_separated = instance.strip().split(delimiter)
                userid = instance_separated[0]
                itemid = instance_separated[1]
                rating = instance_separated[2]

                # map userid and itemid, and update the dictionary
                if userid not in self.userid_mapped_data.keys():
                    self.userid_mapped_data[userid] = user_index
                    user_index += 1
                if itemid not in self.itemid_mapped_data.keys():
                    self.itemid_mapped_data[itemid] = item_index
                    item_index += 1

                data[self.userid_mapped_data[userid], self.itemid_mapped_data[itemid]] = rating

        return data
    def read_train_test_text_file(self, train_path, test_path, train_delimiter, test_delimiter):
        train_df = pd.read_csv(train_path, names=['userid', 'itemid', 'rating'], sep='\t')
        test_df = pd.read_csv(test_path, names=['userid', 'itemid', 'rating'], sep='\t')

        n_users = len(set(list(train_df.userid.unique()) + list(test_df.userid.unique())))
        # n_items = len(set(train_df.itemid.unique()) & set(test_df.itemid.unique()))
        n_items = len(set(list(train_df.itemid.unique()) + list(test_df.itemid.unique())))

        train_data = np.zeros((n_users, n_items))
        with open(train_path) as f:
            for instance in tqdm(f):
                instance_separated = instance.strip().split(train_delimiter)
                userid = instance_separated[0]
                itemid = instance_separated[1]
                rating = instance_separated[2]

                if userid not in self.userid_mapped_data.keys():
                    self.userid_mapped_data[userid] = len(self.userid_mapped_data)
                if itemid not in self.itemid_mapped_data.keys():
                    self.itemid_mapped_data[itemid] = len(self.itemid_mapped_data)

                train_data[self.userid_mapped_data[userid], self.itemid_mapped_data[itemid]] = rating
        print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": train loaded.")

        test_data = np.zeros((n_users, n_items))
        with open(test_path) as f:
            for instance in tqdm(f):
                instance_separated = instance.strip().split(test_delimiter)
                userid = instance_separated[0]
                itemid = instance_separated[1]
                rating = instance_separated[2]

                if userid not in self.userid_mapped_data.keys():
                    self.userid_mapped_data[userid] = len(self.userid_mapped_data)
                if itemid not in self.itemid_mapped_data.keys():
                    self.itemid_mapped_data[itemid] = len(self.itemid_mapped_data)

                test_data[self.userid_mapped_data[userid], self.itemid_mapped_data[itemid]] = rating
        print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": test loaded.")
        return train_data, test_data
    # def read_test_text_file(self, file_path, delimiter):
    #     data = np.zeros((self.train_data.shape))
    #     user_index, item_index = 0, 0
    #     with open(file_path) as f:
    #         for instance in f:
    #             instance_separated = instance.strip().split(delimiter)
    #             userid = instance_separated[0]
    #             itemid = instance_separated[1]
    #             rating = instance_separated[2]
    #
    #             # map userid and itemid, and update the dictionary
    #             # if userid not in self.userid_mapped_data.keys():
    #             #     self.userid_mapped_data[userid] = user_index
    #             #     user_index += 1
    #             # if itemid not in self.itemid_mapped_data.keys():
    #             #     self.itemid_mapped_data[itemid] = item_index
    #             #     item_index += 1
    #
    #             # update the data by appending the new data instance to the data matrix
    #             # if (data.shape[0] - 1) < self.userid_mapped_data[userid]:
    #             #     data = np.append(data, np.zeros((1, data.shape[1])), axis=0)
    #             # if (data.shape[1] - 1) < self.itemid_mapped_data[itemid]:
    #             #     data = np.append(data, np.zeros((1, data.shape[0])).transpose(), axis=1)
    #
    #             if itemid not in self.itemid_mapped_data.keys():
    #                 self.itemid_mapped_data[itemid] = len(self.itemid_mapped_data)
    #             data[self.userid_mapped_data[userid], self.itemid_mapped_data[itemid]] = rating
    #
    #     return data
    def read_user_text_file(self, file_path, delimiter):
        data = np.array(np.zeros((len(self.userid_mapped_data.keys()), 1)))
        feature_index = 0
        with open(file_path) as f:
            for instance in f:
                instance_separated = instance.strip().split(delimiter)
                userid = instance_separated[0]
                feature = instance_separated[1]

                # map user feature, and update the dictionary
                if feature not in self.user_attribute_mapped_data.keys():
                    self.user_attribute_mapped_data[feature] = feature_index
                    feature_index += 1

                # update the data by appending the new data instance to the data matrix
                if (data.shape[1] - 1) < self.user_attribute_mapped_data[feature]:
                    data = np.append(data, np.zeros((1, data.shape[0])).transpose(), axis=1)

                data[self.userid_mapped_data[userid], self.user_attribute_mapped_data[feature]] = 1

        return data
    def read_item_text_file(self, file_path, delimiter, category_column, category_delimiter):
        # file format: itemid, categories
        data = np.array(np.zeros((len(self.itemid_mapped_data.keys()), 1)))
        category_index = 0
        with open(file_path) as f:
            for instance in tqdm(f):
                instance_separated = instance.strip().split(delimiter)
                itemid = instance_separated[0]
                categories = instance_separated[category_column].split(category_delimiter)

                for category in categories:
                    # map categories, and update the dictionary
                    if category not in self.item_topic_mapped_data.keys():
                        self.item_topic_mapped_data[category] = category_index
                        category_index += 1

                    # update the data by appending the new data instance to the data matrix
                    if (data.shape[1] - 1) < self.item_topic_mapped_data[category]:
                        data = np.append(data, np.zeros((1, data.shape[0])).transpose(), axis=1)
                    if itemid in self.itemid_mapped_data.keys():
                        data[self.itemid_mapped_data[itemid], self.item_topic_mapped_data[category]] = 1

        # Adding category "unknown" to the dictionary to be assigned to the items with no category
        col_sum = data.sum(axis=1)
        if data[col_sum == 0].size != 0:
            self.item_topic_mapped_data['unknown'] = category_index
            data = np.append(data, np.zeros((1, data.shape[0])).transpose(), axis=1)
            data[col_sum == 0, category_index] = 1 # equivalent to data[col_sum == 0, -1] = 1

        return data
    def read_supplier_text_file(self, file_path, delimiter):
        # file format: supplierid, itemid
        data = np.array(np.zeros((len(self.itemid_mapped_data.keys()), 1)))
        supplier_index = 0
        with open(file_path) as f:
            for instance in f:
                instance_separated = instance.strip().split(delimiter)
                itemid = instance_separated[1]
                supplierid = instance_separated[0]

                # map categories, and update the dictionary
                if supplierid not in self.supplierid_mapped_data.keys():
                    self.supplierid_mapped_data[supplierid] = supplier_index
                    supplier_index += 1

                # update the data by appending the new data instance to the data matrix
                if (data.shape[1] - 1) < self.supplierid_mapped_data[supplierid]:
                    data = np.append(data, np.zeros((1, data.shape[0])).transpose(), axis=1)

                if itemid in self.itemid_mapped_data.keys():
                    data[self.itemid_mapped_data[itemid], self.supplierid_mapped_data[supplierid]] = 1

        return data[:, np.where(sum(data) > 0)[0]]

    def load_data_from_text_file(self, config):
        self.ratings, self.train_data, self.test_data, self.user_data, self.item_data, self.supplier_data = None, None, None, None, None, None
        if self.rating_file_path is not None:
            n_rows, n_cols = config.get_rating_data_size()
            if n_rows is not None and n_cols is not None:
                self.ratings = self.read_rating_text_file_given_size(self.rating_file_path, config.get_rating_data_delimiter(), n_rows, n_cols)
            else:
                self.ratings = self.read_rating_text_file(self.rating_file_path, config.get_rating_data_delimiter())
        if config.get_splitter_model() == 'giventraintest':
            # self.train_data = self.read_rating_text_file(self.train_file_path, config.get_train_data_delimiter())
            # self.test_data = self.read_test_text_file(self.test_file_path, config.get_test_data_delimiter())
            print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Start loading train/test.")
            self.train_data, self.test_data = self.read_train_test_text_file(self.train_file_path, self.test_file_path, config.get_train_data_delimiter(), config.get_test_data_delimiter())
            print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": train/test loaded.")

        if self.user_file_path is not None:
            self.user_data = self.read_user_text_file(self.user_file_path, config.get_user_data_delimiter())
        if self.item_file_path is not None:
            print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Start loading items.")
            self.item_data = self.read_item_text_file(self.item_file_path, config.get_item_data_delimiter(), config.get_item_data_category_column(), config.get_item_data_category_delimiter())
            print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Items loaded.")
        if self.supplier_file_path is not None:
            self.supplier_data = self.read_supplier_text_file(self.supplier_file_path, config.get_supplier_data_delimiter())

        self.sample_rating_by_active_user_item(config)
        self.binarize_rating_data(config)
        if config.get_splitter_model() == 'splittraintest':
            self.train_data, self.test_data = self.split_rating_data(self.ratings, config.get_splitter_model_traintest_ratio(), config.seed)

        self.save_data_as_txt_file(config)

        print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Data loaded.")

    def sample_rating_by_active_user_item(self, config):
        # Get the most active users and most popular items
        if config.get_sampling_model() == 'active_user_item':
            F_idx = np.argsort(np.count_nonzero(self.ratings, axis=0))[-config.max_item:]
            self.ratings = self.ratings[:, F_idx]
            if hasattr(self, 'item_data'):
                if self.item_data is not None:
                    self.item_data = self.item_data[F_idx, :]
            if hasattr(self, 'supplier_data'):
                if self.supplier_data is not None:
                    self.supplier_data = self.supplier_data[F_idx, :]
                    self.supplier_data = self.supplier_data[:, np.where(sum(self.supplier_data) > 0)[0]]
            u_idx = np.argsort(np.count_nonzero(self.ratings, axis=1))[-config.max_user:]
            self.ratings = self.ratings[u_idx, :]
            if hasattr(self, 'user_data'):
                if self.user_data is not None:
                    self.user_data = self.user_data[u_idx, :]
        elif config.get_sampling_model() == 'active_user':
            u_idx = np.argsort(np.count_nonzero(self.ratings, axis=1))[-config.max_user:]
            self.ratings = self.ratings[u_idx, :]
            if hasattr(self, 'user_data'):
                if self.user_data is not None:
                    self.user_data = self.user_data[u_idx, :]
    def binarize_rating_data(self, config):
        binarize, threshold = config.get_data_binarization()
        if binarize:
            for u in range(self.ratings.shape[0]):
                for i in range(self.ratings.shape[1]):
                    self.ratings[u, i] = int(self.ratings[u, i]) > threshold
    def split_rating_data(self, rating_data, ratio, seed):
        train_data = np.zeros(rating_data.shape)
        test_data = np.zeros(rating_data.shape)
        np.random.seed(seed=seed)
        for u in range(rating_data.shape[0]):
            u_vec = rating_data[u]
            index_ones = np.asarray(np.where(u_vec != 0))[0]
            np.random.shuffle(index_ones)
            train_size = int(len(index_ones) * ratio)
            test_data[u][index_ones[:train_size]] = rating_data[u][index_ones[:train_size]]
            train_data[u][index_ones[train_size:]] = rating_data[u][index_ones[train_size:]]
        return train_data, test_data
    def save_data_as_txt_file(self, config):
        save_data, train_file_name, test_file_name = config.get_save_txt_data_info()
        if save_data:
            f = open(config.get_data_path() / train_file_name, "w")
            for u in tqdm(range(self.train_data.shape[0])):
                # items = np.asarray(np.where(self.train_data[u] == 1))[0]
                items = np.asarray(np.where(self.train_data[u] > 0))[0]
                for j in range(len(items)):
                    f.write(str(list(self.userid_mapped_data.keys())[list(self.userid_mapped_data.values()).index(u)]) + "\t" + str(list(self.itemid_mapped_data.keys())[list(self.itemid_mapped_data.values()).index(items[j])]) + "\t" + str(self.train_data[u, items[j]]) + "\n")
            f.close()

            f = open(config.get_data_path() / test_file_name, "w")
            for u in tqdm(range(self.test_data.shape[0])):
                items = np.asarray(np.where(self.test_data[u] > 0))[0]
                # items = np.asarray(np.where(self.test_data[u] == 1))[0]
                for j in range(len(items)):
                    f.write(str(list(self.userid_mapped_data.keys())[list(self.userid_mapped_data.values()).index(u)]) + "\t" + str(list(self.itemid_mapped_data.keys())[list(self.itemid_mapped_data.values()).index(items[j])]) + "\t" + str(self.test_data[u, items[j]]) + "\n")
            f.close()
    def save_data_as_pkl_file(self, config):
        save_data, file_name = config.get_save_pkl_data_info()
        if save_data:
            save_obj = {'R': self.ratings, 'U': self.user_data, 'I': self.item_data, 'S': self.supplier_data}
            with open(config.get_data_path() / file_name, 'wb') as f:
                pk.dump(save_obj, f)
#
#
#
#
#     config = ConfigCmd(config_file, target, log_filename)
#     print(target)
#     if config.is_valid():
#         # config.load_libraries()
#         config.process_config()
#     # else:
#     #     raise InvalidConfiguration("Configuration file load error", "There was an error loading the configuration file.")
#     return config

    def get_topical_features(self):
        def __w(F, G):
            w = np.zeros(G.shape)
            FG = np.dot(F, G)
            # total number of clicks received by an item
            FG[FG > 1] = 1
            w_den = np.sum(FG, axis=0)
            # total number of clicks received by a topic
            w_num = np.sum(F, axis=0)

            for i in range(G.shape[0]):
                j = np.where(G[i])[0]
                try:
                    w[i, j] = w_num[i] / w_den[j]
                except RuntimeWarning:
                    w[i, j] = 0
            # Now compute theta
            FG = np.dot(F, G)
            theta = np.zeros((F.shape[0], G.shape[1]))
            # numerator is actually FG
            for i in range(theta.shape[0]):
                theta[i] = FG[i] / (np.sum(FG[i]) + 1e-10)
            return w, theta

        train_item_topical_features, train_user_topical_features = __w(self.train_data, self.item_data)
        # train_item_topical_features = train_item_topical_features / np.linalg.norm(train_item_topical_features, axis=1)[:, np.newaxis]
        # train_user_topical_features = train_user_topical_features / np.linalg.norm(train_user_topical_features, axis=1)[:, np.newaxis]
        test_item_topical_features, test_user_topical_features = __w(self.test_data, self.item_data)
        # test_item_topical_features = test_item_topical_features / np.linalg.norm(test_item_topical_features, axis=1)[:, np.newaxis]
        # test_user_topical_features = test_user_topical_features / np.linalg.norm(test_user_topical_features, axis=1)[:, np.newaxis]

        return {"train_item_topical_features": train_item_topical_features, "train_user_topical_features": train_user_topical_features, "test_item_topical_features": test_item_topical_features, "test_user_topical_features": test_user_topical_features}
    def get_latent_features_by_SVD(self, latent_feature_dim, latent_feature_normalize):
        """
        Get the feature and mu from MovieLense based on SVD.
        F = USV, features: SV from training, mu U from testing
        :param f_name: file name
        :param train_percentage:
        :param max_user:
        :param max_item:
        :param d: number of eigenvalue
        :param seed:
        :return:
        """

        u, s, v = np.linalg.svd(self.train_data)
        train_item_features = v[:latent_feature_dim, :].T * s[:latent_feature_dim]
        if latent_feature_normalize:
            train_item_features = train_item_features / (np.linalg.norm(train_item_features, axis=1)[:, np.newaxis] + 1e-10)
            train_item_features = train_item_features / np.sqrt(2)
        train_user_features = u[:, :latent_feature_dim]

        u, s, v = np.linalg.svd(self.test_data)
        test_item_features = v[:latent_feature_dim, :].T * s[:latent_feature_dim]
        if latent_feature_normalize:
            test_item_features = test_item_features / (np.linalg.norm(test_item_features, axis=1)[:, np.newaxis] + 1e-10)
            test_item_features = test_item_features / np.sqrt(2)
        test_user_features = u[:, :latent_feature_dim]

        return {"train_item_latent_features": train_item_features, "train_user_latent_features": train_user_features,
                "test_item_latent_features": test_item_features, "test_user_latent_features": test_user_features}
    def get_latent_features_by_SVD_LeastSquare(self, latent_feature_dim, latent_feature_normalize):
        """
        Get the feature and mu from MovieLense based on SVD.
        F = USV, features: SV from training, mu U from testing
        :param f_name: file name
        :param train_percentage:
        :param max_user:
        :param max_item:
        :param d: number of eigenvalue
        :param seed:
        :return:
        """

        u, s, v = np.linalg.svd(self.train_data)
        # u, s, v = np.linalg.svd(self.train_data, full_matrices=False)
        # u, s, v = svds(self.train_data, k=10)
        train_item_features = v[:latent_feature_dim, :].T * s[:latent_feature_dim]
        if latent_feature_normalize:
            train_item_features = train_item_features / (np.linalg.norm(train_item_features, axis=1)[:, np.newaxis] + 1e-10)
            train_item_features = train_item_features / np.sqrt(2)
        # Y = F_test
        #### Y = F_test.sum(axis=0) / F_test.shape[0]

        """
        attraction_prob = item_feature.T * user_feature
        user_feature = item_covariance_matrix_inverse * (item_feature * attraction_prob)
        attraction_prob: test_data 
        """
        item_covariance_matrix_inverse = np.linalg.inv(train_item_features.T.dot(train_item_features))
        B = train_item_features.T.dot(self.test_data.T)
        test_user_features = np.dot(item_covariance_matrix_inverse, B)

        return {"train_item_latent_features": train_item_features, "test_user_latent_features": test_user_features.T}

    def load_features(self, config):
        data_features = {}
        if self.train_data is not None and self.test_data is not None:
            # data_features = self.get_latent_features_by_SVD_LeastSquare(config.get_latent_feature_dim(), config.normalize_latent_feature())
            data_features = self.get_latent_features_by_SVD(config.get_latent_feature_dim(), config.normalize_latent_feature())
        if hasattr(self, 'item_data'):
            if self.item_data is not None:
                topical_features = self.get_topical_features()
                data_features = dict(list(topical_features.items()) + list(data_features.items()))
        print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Feature data created.")
        return data_features
        # data_features = {}
        # derivation_type = config.get_features_derivation_type()
        # if 'latent' in derivation_type:
        #     data_features = self.get_latent_features_by_SVD(config.get_latent_feature_dim())
        # if 'topical' in derivation_type:
        #     topical_features = self.get_topical_features()
        #     data_features = dict(list(topical_features.items()) + list(data_features.items()))
        #
        # return data_features
    def save_item_features(self, config):
        if config.save_item_features():
            with open(config.get_data_path() / 'train_item_latent_features.txt', 'w') as f:
                for i in range(self.feature_data['train_item_latent_features'].shape[0]):
                    f.write(str(get_dict_key_from_value(self.itemid_mapped_data, i)) + "\t" + ",".join(map(str, self.feature_data['train_item_latent_features'][i,:])) + "\n")
            if 'train_item_latent_features' in self.feature_data:
                with open(config.get_data_path() / 'train_item_topical_features.txt', 'w') as f:
                    for i in range(self.feature_data['train_item_topical_features'].shape[0]):
                        f.write(str(get_dict_key_from_value(self.itemid_mapped_data, i)) + "\t" + ",".join(map(str, self.feature_data['train_item_topical_features'][i, :])) + "\n")


import numpy as np
from .abstract_ranker import AbstractRanker

# A simple baseline that randomly recommends n_recos playlists to each user.
class Random(AbstractRanker):
    def __init__(self, config, dataObj, parameters=None):
        super(Random, self).__init__(config, dataObj)

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        n_users = len(batch_users)
        rankings = np.zeros((n_users, self.config.list_size), dtype=np.int64)
        r = np.arange(self.dataObj.n_items)
        for i in range(n_users):
            np.random.shuffle(r)
            rankings[i] = r[:self.config.list_size]
        return rankings

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):
        return


import numpy as np
from .abstract_ranker import AbstractRanker
from math import log
from collections import defaultdict

#  Upper Confidence Bound (UCB) strategy, using KL-UCB bounds [Garivier and Cappe, 2011] tailored for Bernoulli rewards
class KLUCB(AbstractRanker):
    def __init__(self, config, dataObj, parameters=None):
        super(KLUCB, self).__init__(config, dataObj)
        # self.user_segment = user_segment
        # n_segments = len(np.unique(self.user_segment))
        self.ranking_display = np.zeros((dataObj.n_users, dataObj.n_items))
        self.ranking_success = np.zeros((dataObj.n_users, dataObj.n_items))
        self.ranking_score = np.ones((dataObj.n_users, dataObj.n_items))
        self.t = 0
        # self.cascade_model = cascade_model
        self.precision = float(parameters["precision"]["value"])
        self.eps = float(parameters["eps"]["value"])

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        # user_segment = np.take(self.user_segment, batch_users)
        user_score = np.take(self.ranking_score, batch_users, axis = 0)
        # Break ties
        user_random_score = np.random.random(user_score.shape)
        user_choice = np.lexsort((user_random_score, -user_score))[:, :self.config.list_size]
        # Shuffle l_init first slots
        # np.random.shuffle(user_choice[0:l_init])
        return user_choice

    def kl(self, x, y):
        x = min(max(x, self.eps), 1 - self.eps)
        y = min(max(y, self.eps), 1 - self.eps)
        return x * log(x / y) + (1 - x) * log((1 - x) / (1 - y))

    def scoring_function(self, n_success, n, t):
        if n == 0:
            return 1.0
        p = n_success / n
        value = p
        u = 1
        threshold = log(t) / n
        _count_iteration = 0
        while _count_iteration < 50 and u - value > self.precision:
            _count_iteration += 1
            m = (value + u) * 0.5
            if self.kl(p, m) > threshold:
                u = m
            else:
                value = m
        return (value + u) * 0.5

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):
        batch_size = len(batch_users)
        modified_data = defaultdict(set)
        for i in range(batch_size):
            # user_segment = self.user_segment[user_ids[i]]
            # total_stream = len(rewards[i].nonzero())
            nb_display = 0
            for r, c in zip(rankings[i], clicks[i]):
                nb_display +=1
                modified_data[batch_users[i]].add(r)
                self.ranking_success[batch_users[i]][r]+=c
                self.ranking_display[batch_users[i]][r]+=1
                # if self.cascade_model and ((total_stream == 0 and nb_display == l_init) or (r == 1)):
                #     break
        self.t = self.ranking_display.sum()
        for seg,pls in modified_data.items():
            for pl in pls:
                self.ranking_score[seg][pl] = self.scoring_function(self.ranking_success[seg][pl], self.ranking_display[seg][pl], self.t)
        return


import numpy as np
from .abstract_ranker import AbstractRanker

# Objective function: min 1/2 (X - U . V_T) + 1/2 * \lambda (U)^2 + 1/2 * \lambda (V)^2

class FactorUCB1(AbstractRanker):
    def __init__(self, config, dataObj, parameters=None):
        super(FactorUCB1, self).__init__(config, dataObj)
        self.prng = np.random.RandomState(seed=config.seed)
        self.k = 0
        self.contextual_var = bool(parameters["contextual_variable"]["value"]) if "contextual_variable" in parameters else False
        if self.contextual_var:
            self.X = self.dataObj.feature_data['train_item_topical_features']
            self.k = self.X.shape[1]
        self.l = int(parameters["latent_dim"]["value"]) if "latent_dim" in parameters else 0
        self.lambda_1 = float(parameters["lambda1"]["value"]) if "lambda1" in parameters else 1.0
        self.lambda_2 = float(parameters["labmda2"]["value"]) if "lambda2" in parameters else 1.0
        self.alpha_a = float(parameters["alpha_a"]["value"]) if "alpha_a" in parameters else 1.0
        self.alpha_u = float(parameters["alpha_u"]["value"]) if "alpha_u" in parameters else 1.0

        self.batch_features = None

        self.U = np.zeros((self.dataObj.n_users, self.k+self.l))
        self.V = np.zeros((self.dataObj.n_items, self.l))

        self.W = np.ones((self.dataObj.n_users, self.dataObj.n_users))

        self.A = self.lambda_1 * np.identity(n=self.k+self.l)
        self.AInv = np.linalg.inv(self.A) #np.zeros((self.k+self.l,self.k+self.l))
        self.b = np.zeros((self.dataObj.n_users,self.k+self.l))

        self.C = self.lambda_2 * np.identity(n=self.l)
        self.CInv = np.linalg.inv(self.C)
        self.d = np.zeros((self.dataObj.n_users, self.l))


    def get_ranking(self, batch_users, sampled_items=None, round=None):
        rankings = np.zeros((len(batch_users), self.config.list_size), dtype=int)
        self.batch_features = np.zeros((len(batch_users), self.config.list_size, self.k+self.l))
        if self.contextual_var:
            items_feature = np.concatenate((self.X, self.V), axis=1)
            user_topic = self.U[:,:self.k]
            user_latent = self.U[:, self.k:]
        else:
            items_feature = self.V
            user_latent = self.U
        tie_breaker = self.prng.rand(len(sampled_items))
        for i in range(len(batch_users)):
            user = batch_users[i]

            XVW = np.dot(items_feature[sampled_items].T,self.W)
            cb1 = self.alpha_u * np.sqrt(np.multiply(np.dot(XVW.T,self.AInv),XVW.T))
            cb2 = self.alpha_a * np.sqrt(np.dot(np.dot(np.dot(user_latent.T, self.W).T, self.CInv), np.dot(user_latent.T, self.W[user]).T))
            score = np.dot(items_feature[sampled_items], np.dot(self.U.T, self.W[user]))
            ucb = score + cb1.T + cb2.T

            rankings[i] = np.lexsort((tie_breaker, -ucb))[:self.config.list_size]
            # if self.contextual_var:
            #     self.batch_features[i] = np.concatenate((self.X[rankings[i]], self.dataObj.feature_data['test_item_latent_features'][rankings[i]]), axis=1)
            # else:
            #     self.batch_features[i] = self.dataObj.feature_data['test_item_latent_features'][rankings[i]]
        return rankings

    def update(self, batch_users, sampled_items, rankings, clicks, round=None, user_round=None):
        for i in range(len(batch_users)):
            user = batch_users[i]

            X = np.zeros((self.V.shape[0], self.k + self.l))
            if self.contextual_var:
                X[rankings[i]] = np.concatenate((self.X[rankings[i]], self.dataObj.feature_data['test_item_latent_features'][rankings[i]]), axis=1)
            else:
                X[rankings[i]] = self.dataObj.feature_data['test_item_latent_features'][rankings[i]]
            X = X[sampled_items]

            _clicks, _X = self.__collect_feedback(clicks, i, rankings, X)

            # discount_coef = [1 / (math.log(1 + j)) for j in range(1, len(rankings[0]) + 1)]
            # discount_coef_reward = [math.log(1 + j) for j in range(1, len(_clicks) + 1)]
            # discount_coef_penalization = [self.gamma * 1 / (math.log(1 + j)) for j in range(1, len(_clicks) + 1)]

            # if self.processing_type == 'recommended_discountfactor':
            #     self.exp_recommended[user][np.array(rankings[0])] += discount_coef
            # elif self.processing_type == 'examined_discountfactor':
            #     if len(clicks) == 0:
            #         self.exp_examined[user][np.array(rankings[0])] += discount_coef
            #     else:
            #         self.exp_examined[user][np.array(rankings[0][:len(clicks)])] += discount_coef[:len(clicks)]
            #
            # if self.processing_type == 'item_weight':
            #     _batch_features = self.update_item_weight(rankings[0], _batch_features, _clicks, discount_coef_penalization, discount_coef_reward, user, user_round)

            cb1 = self.alpha_u * np.sqrt(np.dot(np.dot(np.dot(items_feature[sampled_items].T, self.W[user].T).T, self.AInv), np.dot(items_feature[sampled_items].T, self.W.T)))
            cb2 = self.alpha_a * np.sqrt(np.dot(np.dot(np.dot(user_latent.T, self.W[user]).T, self.CInv), np.dot(user_latent.T, self.W[user])))
            """
            This is for computing self.theta (Line 3-5 of Alogirthm 1 of NIPS 11)
            For fast matrix inverse, we use Woodbury matrix identity (https://en.wikipedia.org/wiki/Woodbury_matrix_identity)

            Return: self.theta is updated.
            """
            # (X,V) * W
            XVW = np.dot(X.T, self.W).T
            # for the inverse of M, feature matrix
            # XW * A^-1 * x^T
            xAx = np.dot(XVW, np.dot(self.AInv, XVW.T))
            # (1/sigma I + xAx)^-1
            try:
                tmp_inv = np.linalg.inv(1 / self.sigma * np.eye(len(XVW)) + xAx)
            except np.linalg.LinAlgError:
                # for the ill matrix. if the matrix is not invertible, we ignore this update
                self.ill_matrix_counter += 1
                return
            # A^-1*x^T
            AInv_xT = self.AInv.dot(XVW.T)
            # AInv_xT*tmp_inv*AInv_xT^T
            self.AInv_tmp = np.dot(np.dot(AInv_xT, tmp_inv), AInv_xT.T)
            # MInv - the new part
            self.AInv -= self.AInv_tmp

            self.A += self.sigma * XVW.T.dot(XVW)

            # for b: feedback
            # if self.processing_type == 'feature_weight':
            #     self.update_feature_weight(_batch_features, _clicks, discount_coef_penalization, discount_coef_reward,
            #                                user, user_round)
            # else:
            self.b += np.dot(_clicks, XVW)

            # self.b_tmp[user] = np.dot(_clicks, _batch_features)
            # self.b[user] += self.b_tmp[user]

            # for parameter theta
            self.theta[user] = np.dot(self.MInv[user], self.b[user])
            # self.theta[self.theta < 0] = 0

            self.n_samples[user] += len(_clicks)
            self.n_clicks[user] += sum(_clicks)

    def __collect_feedback(self, clicks, batch_user_id, rankings, batch_X):
        """
        :param y:
        :return: the last observed position.
        """
        # With  Cascade assumption, only the first click counts.
        if self.config.feedback_model == 'cascade':
            if np.sum(clicks[batch_user_id]) == 0:
                return clicks[batch_user_id], batch_X
            first_click = np.where(clicks[batch_user_id])[0][0]
            batch_X[rankings[:first_click+1]] = np.zeros((rankings[:first_click+1], batch_X.shape[1]))
            return clicks[batch_user_id][:first_click + 1], self.batch_features[batch_user_id][:first_click + 1]
        elif self.config.feedback_model == 'dcm':
            if np.sum(clicks[batch_user_id]) == 0:
                return clicks[batch_user_id], self.batch_features[batch_user_id]
            last_click = np.where(clicks[batch_user_id])[0][-1]
            return clicks[batch_user_id][:last_click + 1], self.batch_features[batch_user_id][:last_click + 1]
        # all items are observed
        else:
            return clicks[batch_user_id], self.batch_features[batch_user_id]

    def vectorize(self, M):
        # temp = []
        # for i in range(M.shape[0]*M.shape[1]):
        # 	temp.append(M.T.item(i))
        # V = np.asarray(temp)
        # return V
        return np.reshape(M.T, M.shape[0] * M.shape[1])


from rankers import CascadeLSB
from rankers import LSBGreedy
from rankers import AbstractRanker
from utils.click_simulators import CascadeSimulator, PBMSimulator, DBMSimulator, DCMSimulator
from utils.my_parser import SimulationArgumentParser
from utils.recommendation_history import save_history

import numpy as np
import pickle as pk
import datetime
import json
import os


CLICKMODEL_MAP = {'CM': CascadeSimulator,
                  'cascademodel': CascadeSimulator,
                  'cm': CascadeSimulator,
                  'PBM': PBMSimulator,
                  'pbm': PBMSimulator,
                  'positionbasemodel': PBMSimulator,
                  'DCM': DCMSimulator,
                  'dcm': DCMSimulator,
                  'DBM': DBMSimulator,
                  'dbm': DBMSimulator,
                  'documentbasemodel': DBMSimulator
                  }


class NaiveSimulator(object):

    name ='NaiveSimulator'

    def __init__(self, dataset, sim_args, q_id):
        """

        :param dataset: dict contains train_x, test_x, theta_star
        :param ClickModel: class: click mode
        :param sim_args: the simulation arguments from utils/my_parser.py
                        we use k: n_position, iteration, seed: random seed, output: output dict
        """
        self.sim_args = sim_args
        self.iteration = sim_args.iteration
        self.k = sim_args.K
        self.prng = np.random.RandomState(seed=sim_args.seed)
        # this one is the input for the online algorithms
        self.train_x = dataset['train_x']
        # the following two define the click model
        self.test_x = dataset['test_x']
        self.theta_star = dataset['theta_star']

        self.n_topics = self.train_x.shape[1]
        self.n_items = self.train_x.shape[0]
        ClickModel = CLICKMODEL_MAP[self.sim_args.ClickModel]
        self.click_simulator = ClickModel(theta_star=self.theta_star, seed=sim_args.seed)
        # find the best ranking based on the testing feature space.
        self.best_ranking, self.best_delta = self.__greedy_search(self.test_x, self.k)

        self.q_id = q_id

    def __greedy_search(self, x, k):
        """
        search for the optimal ranking
        :param x: input feature
        :param k: number of postions
        :return: optimal ranking
        """
        delta_t = []
        coverage = np.zeros(self.n_topics)
        ranking = []
        ranking_set = set()
        for i in range(k):
            tie_breaker = self.prng.rand(len(x))
            # Line 8 - 11 of Nips 11
            delta = AbstractRanker.conditional_coverage(x=x, coverage=coverage)
            score = np.dot(delta, self.theta_star)
            tmp_rank = np.lexsort((tie_breaker, -score))
            for tr in tmp_rank:
                if tr not in ranking_set:
                    ranking.append(tr)
                    ranking_set.add(tr)
                    delta_t.append(delta[tr])
                    break
            coverage = AbstractRanker.ranking_coverage(x[ranking])
        return ranking, np.asarray(delta_t)

    def __convert_to_topic_coverage(self, x):
        k, d = x.shape
        delta_t = []
        coverage = np.zeros(d)
        for idx, topic in enumerate(x):
            delta = AbstractRanker.conditional_coverage(x=topic, coverage=coverage)
            delta_t.append((delta))
            coverage = AbstractRanker.ranking_coverage(x[:idx+1])
        return np.asarray(delta_t)

    def run(self, rankers, save_results=False):
        if type(rankers) is not list:
            rankers = [rankers]

        regret = {}
        reward = {}
        for ranker in rankers:
            regret[ranker.name] = np.zeros(self.iteration)
            reward[ranker.name] = np.zeros(self.iteration)

        for i in range(self.iteration):
            if self.sim_args.same_coins:
                self.click_simulator.set_coins(self.k)

            best_clicks, best_reward = self.click_simulator.get_feedback(self.best_delta)
            for ranker in rankers:
                ranking, delta = ranker.get_ranking(self.train_x, self.k, i)
                ranker.n_recommended[np.array(ranking)] = ranker.n_recommended[np.array(ranking)] + 1
                delta = self.__convert_to_topic_coverage(self.test_x[ranking])
                """
                The click is from the click simulator. So it is defined by the testing part. w_test and theta_test.
                Bug from the click log, the bias from train_x to test_x is large. Here, I still use train.
                """
                clicks, t_reward = self.click_simulator.get_feedback(delta)

                if (sum(clicks) > 0):
                    save_history(i, self.q_id, ranking, np.array(ranking)[clicks])
                else:
                    save_history(i, self.q_id, ranking, [])

                ranker.update(y=clicks)
                reward[ranker.name][i] = t_reward
                regret[ranker.name][i] = best_reward - t_reward

            if self.sim_args.same_coins:
                self.click_simulator.del_coins()
        # if save_results:
        #     self.__save_results(rankers=rankers, reward=reward, regret=regret)

        return reward, regret

    def save_results(self, q_id, rankers, reward, regret):
        """
        save results to the self.sim_args.output director
        the name is ranker.name+ranker parameters + random seed + save date
        :param rankers: same as self.run()
        :param reward: output of self.run()
        :param regret: output of self.run()
        :return: save results to json file.
        """
        # Saving directory
        if self.sim_args.output[-1] == '/':
            prefix = self.sim_args.output + \
                     '/'.join([self.sim_args.data_name, self.sim_args.ClickModel, 'norm-'+str(self.sim_args.normalized),
                               'rep'+str(self.sim_args.iteration),
                               'pos'+str(self.sim_args.K), 'topic'+str(self.sim_args.n_topic)]) + '/' + str(q_id) + '/'
        else:
            prefix = self.sim_args.output + '/' + \
                     '/'.join([self.sim_args.data_name, self.sim_args.ClickModel, 'norm-'+str(self.sim_args.normalized),
                               'rep'+str(self.sim_args.iteration),
                               'pos'+str(self.sim_args.K), 'topic'+str(self.sim_args.n_topic)]) + '/' + str(q_id) + '/'

        if not os.path.exists(prefix):
            os.makedirs(prefix)

        suffix = 'seed-' + str(self.sim_args.seed) + \
                 '-' + str(datetime.datetime.now().date()) + \
                 '-' + str(datetime.datetime.now().time())[:8].replace(':', '-') \
                 + '.js'

        for ranker in rankers:
            save_name = prefix + ranker.name + '-alpha%.2f-sigma%.2f-' % (ranker.alpha, ranker.sigma) + suffix
            objs = {'reward': reward[ranker.name].tolist(),
                    'regret': regret[ranker.name].tolist()
                    }
            with open(save_name, 'w') as f:
                json.dump(objs, f)


if __name__ == '__main__':
    args = SimulationArgumentParser()
    sim_args = args.parse_args('-K 2 --output results/ --data synthetic --seed 2'.split())

    SimulationArgumentParser.print(sim_args)

    with open('../data/synthetic.pkl', 'rb') as f:
        dataset = pk.load(f)

    d = dataset['train_x'].shape[1]
    rankers = [CascadeLSB(d=d, sigma=.1, alpha=.8, seed=sim_args.seed),
               LSBGreedy(d=d, sigma=.1, alpha=.8, seed=sim_args.seed)
              ]

    sim = OnlineSimulator(dataset=dataset, sim_args=sim_args)
    reward, regret = sim.run(rankers=rankers, save_results=True)

    for key in reward:
        print(key, (np.cumsum(reward[key])/np.arange(1, 1+sim_args.iteration))[-10:])

    for key in regret:
        print(key, np.cumsum(regret[key])[-10:])




from pathlib import Path


class Files:
    """
    This class encapsulates the file system for librec-auto.

    The files for librec-auto can be stored in a number of possible locations:
    - global location: determined by the install location of librec-auto
    - rules-specific location: determined by the location of a rules directory [Not implemented]
    - user-specific location: associated with the user's home directory [Not implemented]
    - experiment-specific location: in a directory associated with a specific experiment.

    These are the types of files that are managed:
    - configuration files (XML). Stores information about how experiments are configured.
    - properties files (key=value format). Generated from configuration files for input to librec
    - split files (UIR or UIRT format). Generated by separating input rules files into different test/training splits
    - experiment logs (csv files). Content is algorithm-specific. Generated by a run of an experiment

    NOTE: User directory is currently unused.
    """

    # _DEFAULT_GLOBAL_DIR_STR = inspect.getfile(fair_dynamic_rec)

    _DEFAULT_CONFIG_DIR_NAME = "conf"
    # _DEFAULT_RES_DIR_NAME = "result"
    # _DEFAULT_DATA_DIR_NAME = "data"
    # _DEFAULT_POST_DIR_NAME = "post"
    # _DEFAULT_LIBRARY_DIR_NAME = "lib"
    # _EXP_DIR_FORMAT = "exp{:05d}"
    # _EXP_DIR_PATTERN = "exp(\d+)"
    # _OUTPUT_FILE_NAME = "output.xml"

    DEFAULT_CONFIG_FILENAME = "config.xml"
    # DEFAULT_REF_EXP_FILENAME = "ref_exp.txt"
    # DEFAULT_TRAIN_FILENAME = "train.txt"
    # DEFAULT_TEST_FILENAME = "test.txt"

    # _DEFAULT_SPLIT_DIR_NAME = "split"

    def __init__(self):
        self._config_dir_path = Path(self._DEFAULT_CONFIG_DIR_NAME)
        # self._data_dir_path = Path(self._DEFAULT_DATA_DIR_NAME)
        # self._jar_dir_path = Path(self._DEFAULT_JAR_DIR_NAME)
        # self._post_dir_path = Path(self._DEFAULT_POST_DIR_NAME)
        # self._lib_dir_path = Path(self._DEFAULT_LIBRARY_DIR_NAME)
        # self._exp_path_dict = OrderedDict()
        #
        #
        # module_init_path = Path(Files._DEFAULT_GLOBAL_DIR_STR).parent
        # self._global_path = module_init_path.parent
        #
        # self._split_dir_path = Path(self._DEFAULT_SPLIT_DIR_NAME)

    # Paths related to the librec installation
    # def get_global_path(self):
    #     return self._global_path
    #
    # def set_global_path(self, path):
    #     self._global_path = Path(path)
    #
    # def get_data_path(self):
    #     return self.get_study_path() / self._data_dir_path
    #
    # def set_data_path(self, config_xml):
    #     data_dir_elem = single_xpath(config_xml, '/librec-auto/data/data-dir')
    #     if data_dir_elem is None:
    #         logging.warning("Configuration file missing data-dir element. Assuming 'data'.")
    #     else:
    #         self._data_dir_path = data_dir_elem.text
    #
    # def get_rules_path(self):
    #     return self.get_global_path() / self._DEFAULT_RULES_FILE
    #
    # def get_lib_path(self):
    #     return self.get_global_path() / self._DEFAULT_LIB_DIR_NAME
    #
    # def get_jar_path(self):
    #     return self.get_global_path() / self._jar_dir_path
    #
    # def get_classpath(self):
    #     return (self.get_jar_path() /
    #             self._DEFAULT_LA_JAR).absolute().as_posix()
    #
    # # Paths related to the current study
    # def get_study_path(self):
    #     return self._study_path
    #
    # def get_status_path(self):
    #     return self._study_path / self._OUTPUT_FILE_NAME

    def get_config_file_path(self):
        return self._study_path / self._config_dir_path / self._config_file_name

    # def get_config_dir_path(self):
    #     return self._study_path / self._config_dir_path
    #
    # def set_post_path(self, path):
    #     self._post_dir_path = Path(path)
    #
    # def get_post_path(self):
    #     return self._study_path / self._post_dir_path
    #
    # def get_split_path(self):
    #     return self._data_dir_path / self._split_dir_path

    def set_study_path(self, path):
        self._study_path = Path(path)

    def set_config_file(self, filename):
        self._config_file_name = Path(filename)



import argparse
import numpy as np
import sys
from fair_dynamic_rec.core.config_cmd import ConfigCmd
from fair_dynamic_rec.core.data_cmd import DataCmd
from pathlib import Path
from fair_dynamic_rec.core import read_config_file
from fair_dynamic_rec.core.util.files import Files
from fair_dynamic_rec.core.util.utils import create_log_name
from fair_dynamic_rec.core.simulators.simulation import OnlineSimulator
from fair_dynamic_rec.core.visualization_cmd import VisualizationCmd
from fair_dynamic_rec.core.metric_cmd import MetricCmd
from datetime import datetime
from fair_dynamic_rec.core.util.ranker_utils import set_rankers
# import multiprocessing as mp

def read_args():
    '''
    Parse command line arguments.
    :return:
    '''
    parser = argparse.ArgumentParser(description= 'The FairDynamicRec tool for running recommender systems experiments in dynamic setting.', epilog='For documentation, refer to: ')

    # todo remove py-eval AS
    # parser.add_argument('action', choices=['run', 'split', 'eval', 'rerank', 'post', 'purge', 'status', 'describe', 'check', 'py-eval'], nargs='?')
    parser.add_argument('action', choices=['run', 'split', 'eval', 'rerank', 'post', 'purge', 'status', 'describe', 'check', 'py-eval'])

    parser.add_argument("-t", "--target", help="Path to experiment directory")#, default='demo01')

    # Optional with arguments
    parser.add_argument("-c", "--conf", help="Use the specified configuration file")

    input_args = parser.parse_args()
    # error_check(vars(input_args))
    return vars(input_args)

def load_config(args: dict) -> ConfigCmd:
    config_file = Files.DEFAULT_CONFIG_FILENAME

    if args['conf']:  # User requested a different configuration file from the default
        config_file = args['conf']

    target = ""
    if (args['target'] != None):
        target = args['target']

    log_file = args['log_name']

    # create a path:

    return read_config_file(config_file, target, log_file)

if __name__ == '__main__':
    print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Running experiment.")

    args = read_args()
    # args['action'] = 'run'
    # args['target'] = '/Users/m.mansouryuva.nl/Research/FairDynamicRec/study'

    # purge_old_logs(args['target'] + "/*")
    log_name = create_log_name('FairDynamicRec_log{}.log')
    args['log_name'] = log_name
    # print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Start loading config.")
    config = load_config(args)
    # print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Config loaded.")

    dataObj = DataCmd(config)
    print("#users: " + str(dataObj.n_users) + ", #items: " + str(dataObj.n_items) + ", #train_ratings:" + str(np.count_nonzero(dataObj.train_data)) + ", #test_ratings:" + str(np.count_nonzero(dataObj.test_data)))
    print("#items: " + str(dataObj.item_data.shape[0]) + ", #cats: " + str(dataObj.item_data.shape[1]) + ", #itemCats:" + str(np.count_nonzero(dataObj.item_data)))
    sys.stdout.flush()

    # print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Start setting rankers.")
    sys.stdout.flush()
    rankers = set_rankers(config, dataObj)
    # print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Rankers set.")
    sys.stdout.flush()

    # pool = mp.Pool(config.processor_count)
    if args['action'] == 'run':
        sim = OnlineSimulator(config, dataObj)
        sim.run(config, dataObj, rankers)

    if args['action'] == 'run' or args['action'] == 'eval':
        viz = VisualizationCmd(config, dataObj, rankers)
        metric = MetricCmd(config, dataObj, rankers)
        metric.compute(config, dataObj, viz)


import numpy as np

class Environment():
    def __init__(self, config, dataObj):
        self.param_U = np.zeros((dataObj.n_items, config.get_latent_feature_dim()))
        self.param_V = np.zeros((dataObj.n_items, config.get_latent_feature_dim()))


import numpy as np


class AbstractSimulator(object):
    name = 'Abstract'

    def __init__(self, seed=42):
        self.prng = np.random.RandomState(seed=seed)

    def get_feedback(self, rank):
        raise NotImplementedError


class CascadeSimulator(AbstractSimulator):
    name = 'CM'

    def __init__(self, theta_star, *args, **kwargs):
        super(CascadeSimulator, self).__init__(*args, **kwargs)
        assert np.linalg.norm(theta_star) <= 1.00001
        self.theta = theta_star
        self.coins = None
        self.coins_ready = False

    def set_coins(self, k):
        """
        -- call this, if you want to use the same coins for each positions each round
        -- set coins for each round.
        -- this should be called before get_feedback
        """
        self.coins = self.prng.rand(k)
        self.coins_ready = True

    def get_feedback(self, delta):
        if type(delta) is list:
            delta = np.asarray(delta)
        delta = np.tile(delta, (1, 1))
        assert delta.shape[1] == self.theta.shape[0]

        if not self.coins_ready:
            self.set_coins(len(delta))
            self.del_coins()
        score = np.dot(delta, self.theta)
        score[score > 1] = 1.
        coins = score >= self.coins
        click = np.where(coins)[0]
        if len(click) > 0:
            coins[click[0]+1:] = False
        return coins, sum(coins)

    def del_coins(self):
        """
        This should be called by simulator at the end of each round.
        :return:
        """
        self.coins_ready = False

    def ave_feedback(self, delta):
        score = np.dot(delta, self.theta)
        score[score > 1] = 1.
        return 1 - np.prod(1-score)


class DCMSimulator(CascadeSimulator):
    name = 'DCM'

    def __init__(self, theta_star, stop_prob=None, *args, **kwargs):
        super(DCMSimulator, self).__init__(theta_star, *args, **kwargs)
        assert np.linalg.norm(theta_star) <= 1
        self.theta = theta_star
        self.coins = None
        self.stop_coins = None
        self.coins_ready = False
        if not stop_prob:
            self.stop_prob = [0.6555304, 0.4868164, 0.46051615, 0.46315161,
                              0.45642676, 0.47130397, 0.50317268, 0.54764235,
                              0.65359742, 0.0]
            # self.stop_prob = 1. / np.arange(1, 11)
        else:
            self.stop_prob = stop_prob

    def set_coins(self, k):
        """
        -- call this, if you want to use the same coins for each positions each round
        -- set coins for each round.
        -- this should be called before get_feedback
        """
        self.coins = self.prng.rand(k)
        self.stop_coins = self.prng.rand(k)
        self.coins_ready = True

    def get_feedback(self, delta):
        if type(delta) is list:
            delta = np.asarray(delta)
        delta = np.tile(delta, (1, 1))
        assert delta.shape[1] == self.theta.shape[0]
        k = len(delta)
        assert k <= len(self.stop_prob)

        if not self.coins_ready:
            self.set_coins(k)
            self.del_coins()
        score = np.dot(delta, self.theta)
        score[score > 1] = 1.
        coins = score >= self.coins
        stops = self.stop_prob[:k] >= self.stop_coins
        stop_click = stops * coins
        if sum(stop_click) == 0:
            stop_pos = len(score)
        else:
            stop_pos = np.flatnonzero(stop_click)[0] + 1

        coins[stop_pos:] = False
        stop_click[stop_pos:] = False
        return coins, sum(stop_click)


class PBMSimulator(AbstractSimulator):
    """
    Position based model
    """
    name = 'PBM'

    def __init__(self, theta_star, examination_prob=None, *args, **kwargs):
        """
        :param theta_star:
        :param examination_prob: examination probability
        :param args:
        :param kwargs:
        """
        super(PBMSimulator, self).__init__(*args, **kwargs)
        assert np.linalg.norm(theta_star) <= 1
        self.theta = theta_star
        if not examination_prob:
            self.exam_prob = 1 / np.arange(1, 11)
            self.exam_prob = np.asarray([0.99999257, 0.96679847, 0.7973465, 0.63112651,
                                         0.50237947, 0.41921298, 0.35512778, 0.30566137,
                                         0.28128806, 0.2852233])
            self.K = 10
        else:
            self.K = len(examination_prob)
            self.exam_prob = examination_prob
        self.coin_pos = None
        self.coin_att = None
        self.coins_ready = False

    def set_coins(self, k):
        """
        -- call this, if you want to use the same coins for each positions each round
        -- set coins for each round.
        -- this should be called before get_feedback
        """
        self.coin_pos = self.prng.rand(self.K)
        self.coin_att = self.prng.rand(k)
        self.coins_ready = True

    def get_feedback(self, delta):
        if type(delta) is list:
            delta = np.asarray(delta)
        delta = np.tile(delta, (1, 1))
        assert delta.shape[1] == self.theta.shape[0]
        assert len(delta) <= self.K

        if not self.coins_ready:
            self.set_coins(len(delta))
            self.del_coins()

        k = len(delta)
        score = np.dot(delta, self.theta)
        score[score > 1] = 1.
        att_coins = score >= self.coin_att
        pos_coins = self.exam_prob >= self.coin_pos
        click = att_coins * pos_coins[:k]

        return click, sum(click)

    def del_coins(self):
        """
        This should be called by simulator at the end of each round.
        :return:
        """
        self.coins_ready = False


class DBMSimulator(CascadeSimulator):
    name = 'Document based model'

    def __init__(self, theta_star, *args, **kwargs):
        super(DBMSimulator, self).__init__(theta_star, *args, **kwargs)

    def get_feedback(self, delta):
        if type(delta) is list:
            delta = np.asarray(delta)
        delta = np.tile(delta, (1, 1))
        assert delta.shape[1] == self.theta.shape[0]

        if not self.coins_ready:
            self.set_coins(len(delta))
            self.del_coins()
        score = np.dot(delta, self.theta)
        score[score > 1] = 1.
        coins = score >= self.coins
        return coins, sum(coins)

class NaiveCascadeSimulator(AbstractSimulator):
    name = 'NCM'

    def __init__(self, *args, **kwargs):
        super(NaiveCascadeSimulator, self).__init__(*args, **kwargs)
        # assert np.linalg.norm(theta_star) <= 1.00001
        # self.theta = theta_star
        self.coins = None
        self.coins_ready = False

    def set_coins(self, k):
        """
        -- call this, if you want to use the same coins for each positions each round
        -- set coins for each round.
        -- this should be called before get_feedback
        """
        self.coins = self.prng.rand(k)
        self.coins_ready = True

    def get_feedback(self, weights):
        # if type(delta) is list:
        #     delta = np.asarray(delta)
        # delta = np.tile(delta, (1, 1))
        # assert delta.shape[1] == self.theta.shape[0]

        if not self.coins_ready:
            self.set_coins(len(weights))
            self.del_coins()
        # score = np.dot(delta, self.theta)
        weights[weights > 1] = 1.
        coins = weights >= self.coins
        click = np.where(coins)[0]
        if len(click) > 0:
            coins[click[0]+1:] = False
        return coins, sum(coins)

    def del_coins(self):
        """
        This should be called by simulator at the end of each round.
        :return:
        """
        self.coins_ready = False

    def ave_feedback(self, delta):
        score = np.dot(delta, self.theta)
        score[score > 1] = 1.
        return 1 - np.prod(1-score)



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="FairDynamicRec",
    version="0.0.77",
	scripts=['fair_dynamic_rec/__main__.py'] ,
    author="Masoud Mansoury",
    author_email="masoodmansoury@gmail.com",
    description=
    "The FairDynamicRec project aims to facilitate recommender system experiments in dynamic setting where system is operating over time.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    # url="https://github.com/that-recsys-lab/librec-auto",
    packages=setuptools.find_packages(),
	include_package_data=True,
	install_requires=['matplotlib',
                      'pandas',
                      'numpy',
                      'progressbar',
                      'lxml',
                      'scipy',
                      'tqdm',
                      'joblib',
                      'sklearn'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



from .ea_linear_submodular_bandit import EALSB
import numpy as np
import math

class EAHybridLSBLinUCBItemWeight(EALSB):
    def __init__(self, config, dataObj, parameters=None):
        """
        The first part of features are for the diversity and the second part is for relevance.
        :param args:
        :param kwargs:
        """
        super(EAHybridLSBLinUCBItemWeight, self).__init__(config, dataObj, parameters)
        self.topic_dim = self.dataObj.feature_data['train_item_topical_features'].shape[1]
        self.latent_dim = self.dataObj.feature_data['train_item_latent_features'].shape[1]

        # for the topic
        self.beta = np.ones((self.dataObj.n_users, self.topic_dim))
        self.b_z = np.zeros((self.dataObj.n_users, self.topic_dim))  # d
        self.A_z = np.zeros((self.dataObj.n_users, self.topic_dim, self.topic_dim))  # d by d
        self.A_z_inv = np.zeros((self.dataObj.n_users, self.topic_dim, self.topic_dim))
        for i in range(self.dataObj.n_users):
            self.A_z[i] = np.eye(self.topic_dim)
            self.A_z_inv[i] = np.eye(self.topic_dim)

        # for the relevance
        self.theta = np.ones((self.dataObj.n_users, self.latent_dim))
        self.b_x = np.zeros((self.dataObj.n_users, self.latent_dim))
        self.A_x = np.zeros((self.dataObj.n_users, self.latent_dim, self.latent_dim))
        self.A_x_inv = np.zeros((self.dataObj.n_users, self.latent_dim, self.latent_dim))
        for i in range(self.dataObj.n_users):
            self.A_x[i] = np.eye(self.latent_dim)
            self.A_x_inv[i] = np.eye(self.latent_dim)
        self.B = np.zeros((self.dataObj.n_users, self.latent_dim, self.topic_dim))

        self.X_topical = np.zeros((self.dataObj.n_users, self.dataObj.feature_data['train_item_topical_features'].shape[0], self.dataObj.feature_data['train_item_topical_features'].shape[1]))
        self.X_latent = np.zeros((self.dataObj.n_users, self.dataObj.feature_data['train_item_latent_features'].shape[0], self.dataObj.feature_data['train_item_latent_features'].shape[1]))
        for i in range(self.dataObj.n_users):
            self.X_topical[i] = self.dataObj.feature_data['train_item_topical_features']
            self.X_latent[i] = self.dataObj.feature_data['train_item_latent_features']
        self.gamma = float(parameters["gamma"]["value"])

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):
        for i in range(len(batch_users)):
            user = batch_users[i]
            _clicks, _batch_features = self.__collect_feedback(clicks, i)

            discount_coef = [pow(self.gamma, j) for j in range(1, len(_clicks) + 1)]
            # 1
            # discount_coef_reward = [1 + pow(self.gamma, len(rankings[i])-j+1) for j in range(1, len(_clicks) + 1)]
            # discount_coef_penalization = [x+1 for x in discount_coef]
            # 2
            # discount_coef_reward = [pow(self.gamma, len(rankings[i]) - j + 1) for j in range(1, len(_clicks) + 1)]
            # discount_coef_penalization = [x for x in discount_coef]
            # 3
            # discount_coef_reward = [pow(self.gamma, len(rankings[i]) - j + 1) for j in range(1, len(_clicks) + 1)]
            # discount_coef_penalization = [pow(1 - self.gamma, j) for j in range(1, len(_clicks) + 1)]
            # # 4
            # discount_coef_reward = [1 + pow(self.gamma, len(rankings[i]) - j + 1) for j in range(1, len(_clicks) + 1)]
            # discount_coef_penalization = [pow(1 - self.gamma, j) for j in range(1, len(_clicks) + 1)]
            # 6
            discount_coef_reward = [1 + math.log(j) for j in range(1, len(_clicks) + 1)]
            # # 7
            # discount_coef_reward = [1 + math.log(j) for j in range(1, len(_clicks) + 1)]
            # discount_coef_penalization = [self.gamma * 1 / (1 + math.log(j)) for j in range(1, len(_clicks) + 1)]

            if self.type == 'item_weight':
                clicked_items_index = np.where(_clicks == 1)[0]
                if len(clicked_items_index) == 0:
                    self.X_topical[user][rankings[i], :] = np.multiply(np.array(discount_coef).reshape(len(_clicks), 1), self.X_topical[user][rankings[i], :])
                    self.X_latent[user][rankings[i], :] = np.multiply(np.array(discount_coef).reshape(len(_clicks), 1), self.X_latent[user][rankings[i], :])
                else:
                    previous_clicked_item_index = 0
                    for clicked_item_index in clicked_items_index:
                        current_clicked_item_index = clicked_item_index
                        # penalizing unclicked items
                        self.X_topical[user][rankings[i][current_clicked_item_index], :] = (2 - discount_coef[current_clicked_item_index]) * self.X_topical[user][rankings[i][current_clicked_item_index], :]
                        self.X_latent[user][rankings[i][current_clicked_item_index], :] = (2 - discount_coef[current_clicked_item_index]) * self.X_latent[user][rankings[i][current_clicked_item_index], :]
                        # rewarding clicked items
                        self.X_topical[user][rankings[i][previous_clicked_item_index: current_clicked_item_index], :] = np.multiply(np.array(discount_coef[previous_clicked_item_index: current_clicked_item_index]).reshape(current_clicked_item_index - previous_clicked_item_index, 1), self.X_topical[user][rankings[i][previous_clicked_item_index: current_clicked_item_index], :])
                        self.X_latent[user][rankings[i][previous_clicked_item_index: current_clicked_item_index], :] = np.multiply(np.array(discount_coef[previous_clicked_item_index: current_clicked_item_index]).reshape(current_clicked_item_index - previous_clicked_item_index, 1), self.X_latent[user][rankings[i][previous_clicked_item_index: current_clicked_item_index], :])
                        previous_clicked_item_index = current_clicked_item_index + 1

            """
            Algorithm 2 of WWW 2010

            Return: self.theta is updated.
            """
            z = _batch_features[:, :self.topic_dim]
            x = _batch_features[:, self.topic_dim:]

            BA = np.matmul(self.B[user].T, self.A_x_inv[user])
            self.A_z[user] += np.matmul(BA, self.B[user])
            self.b_z[user] += np.dot(BA, self.b_x[user])

            self.A_x[user] += np.dot(x.T, x)
            self.B[user] += np.dot(x.T, z)
            if self.type == 'feature_weight':
                clicked_items_index = np.where(_clicks == 1)[0]
                _x = x
                # if len(clicked_items_index) == 0:
                #     _x = -np.multiply(np.array(discount_coef_penalization).reshape(len(_clicks), 1), _x)
                # else:
                #     previous_clicked_item_index = 0
                for clicked_item_index in clicked_items_index:
                    current_clicked_item_index = clicked_item_index
                    _x[current_clicked_item_index, :] = discount_coef_reward[current_clicked_item_index] * _x[current_clicked_item_index, :]
                        # _x[previous_clicked_item_index: current_clicked_item_index, :] = -np.multiply(np.array(discount_coef_penalization[previous_clicked_item_index: current_clicked_item_index]).reshape(current_clicked_item_index-previous_clicked_item_index,1), _x[previous_clicked_item_index: current_clicked_item_index, :])
                        # previous_clicked_item_index = current_clicked_item_index + 1
                # self.b_x[user] += _x.sum(axis=0)
            # else:
            self.b_x[user] += np.dot(_clicks, _x)

            try:
                self.A_x_inv[user] = np.linalg.inv(self.A_x[user])
            except:
                self.A_x_inv[user] = np.linalg.pinv(self.A_x[user])

            BA = np.matmul(self.B[user].T, self.A_x_inv[user])
            self.A_z[user] += np.dot(z.T, z) - np.matmul(BA, self.B[user])
            if self.type == 'feature_weight':
                clicked_items_index = np.where(_clicks == 1)[0]
                _z = z
                # if len(clicked_items_index) == 0:
                #     _z = -np.multiply(np.array(discount_coef_penalization).reshape(len(_clicks), 1), _z)
                # else:
                #     previous_clicked_item_index = 0
                for clicked_item_index in clicked_items_index:
                    current_clicked_item_index = clicked_item_index
                    _z[current_clicked_item_index, :] = discount_coef_reward[current_clicked_item_index] * _z[current_clicked_item_index, :]
                        # _z[previous_clicked_item_index: current_clicked_item_index, :] = -np.multiply(np.array(discount_coef_penalization[previous_clicked_item_index: current_clicked_item_index]).reshape(current_clicked_item_index-previous_clicked_item_index,1), _z[previous_clicked_item_index: current_clicked_item_index, :])
                        # previous_clicked_item_index = current_clicked_item_index + 1
                # self.b_z[user] += _z.sum(axis=0) - np.dot(BA, self.b_x[user])
            # else:
            self.b_z[user] += np.dot(_clicks, _z) - np.dot(BA, self.b_x[user])

            try:
                self.A_z_inv[user] = np.linalg.inv(self.A_z[user])
            except:
                self.A_z_inv[user] = np.linalg.pinv(self.A_z[user])

            self.beta[user] = np.dot(self.A_z_inv[user], self.b_z[user])
            B_tmp = self.b_x[user] - np.dot(self.B[user], self.beta[user])
            self.theta[user] = np.dot(self.A_x_inv[user], B_tmp)

            self.n_samples[user] += len(_clicks)
            self.n_clicks[user] += sum(_clicks)

    # def __collect_feedback(self, y):
    #     """
    #     With Cascade assumption, only the first click counts.
    #     :param y: click feedback
    #     :return: position of first click
    #     """
    #     if np.sum(y) == 0:
    #         return len(y)
    #     first_click = np.where(y)[0][0]
    #
    #     return first_click + 1

    # def update(self, y, delta=None):
    #     if delta is None:
    #         delta = self.delta_t
    #     feedback_len = self.__collect_feedback(y=y)
    #     delta = delta[:feedback_len].reshape((feedback_len, self.d))  # make sure it is a matrix
    #     self.__compute_parameters(delta=delta, y=y[:feedback_len])
    #     self.n_samples += len(y)
    #     self.n_clicks += sum(y)

    def __collect_feedback(self, clicks, batch_user_id):
        """
        :param y:
        :return: the last observed position.
        """
        # With  Cascade assumption, only the first click counts.
        if self.config.feedback_model == 'cascade':
            if np.sum(clicks[batch_user_id]) == 0:
                return clicks[batch_user_id], self.batch_features[batch_user_id]
            first_click = np.where(clicks[batch_user_id])[0][0]
            return clicks[batch_user_id][:first_click + 1], self.batch_features[batch_user_id][:first_click + 1]
        elif self.config.feedback_model == 'dcm':
            if np.sum(clicks[batch_user_id]) == 0:
                return clicks[batch_user_id], self.batch_features[batch_user_id]
            last_click = np.where(clicks[batch_user_id])[0][-1]
            return clicks[batch_user_id][:last_click + 1], self.batch_features[batch_user_id][:last_click + 1]
        # all items are observed
        else:
            return clicks[batch_user_id], self.batch_features[batch_user_id]

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        """
        :param x: n * (n_topic + n_feature) np array
        :param k: number of positions
        :return: ranking: the ranked item id.
        """
        # assert x.shape[0] >= k
        rankings = np.zeros((len(batch_users), self.config.list_size), dtype=int)
        self.batch_features = np.zeros((len(batch_users), self.config.list_size, self.topic_dim + self.latent_dim))
        tie_breaker = self.prng.rand(len(self.dataObj.feature_data['train_item_latent_features']))

        # self.delta_t = np.zeros((k, x.shape[1]))
        # delta = x

        # z = delta[:, :self.n_z]  # topic
        # x = delta[:, self.n_z:]  # feature

        # tie_breaker = self.prng.rand(len(delta))
        for i in range(len(batch_users)):
            user = batch_users[i]
            # for CB
            BA_X = np.matmul(self.B[user].T, self.A_x_inv[user])  # B.T  A_x^-1
            ABA = np.matmul(self.A_z_inv[user], BA_X)  # A_z^-1 B.T A_X^-1
            ABABA = np.matmul(BA_X.T, ABA)

            # score and cb for the relevance
            score_x = np.dot(self.X_latent[user], self.theta[user])
            XAX = np.multiply(np.dot(self.X_latent[user], self.A_x_inv[user]), self.X_latent[user]).sum(axis=1)  # x^T A_X^-1 X^T
            cb_x = np.multiply(np.dot(self.X_latent[user], ABABA), self.X_latent[user]).sum(axis=1) + XAX
            ucb_x = score_x + 1e-6 * tie_breaker
            ABAX = np.dot(ABA, self.X_latent[user].T).T
            # score and cb for the topic
            # delta_t = []
            batch_features = []
            coverage = np.zeros(self.topic_dim)
            ranking = []
            # ranking_set = set()
            for j in range(self.config.list_size):
                # Line 8 - 11 of Nips 11
                z_t = self.conditional_coverage(x=self.X_topical[user], coverage=coverage)
                ZAZ = np.multiply(np.dot(z_t, self.A_z_inv[user]), z_t).sum(axis=1)  # Z^T A_Z^-1 Z^T
                # cb_z = ZAZ - 2 * np.multiply(np.dot(z_t, ABA), x).sum(axis=1)
                cb_z = ZAZ - 2 * np.multiply(z_t, ABAX).sum(axis=1)

                # if (self.mitigation is None):
                cb = self.alpha * np.sqrt(cb_z + cb_x)
                # else:
                #     cb = self.alpha * (1 - (self.n_recommended / (t + 1))) * np.sqrt(cb_z + cb_x)
                score_z = np.dot(z_t, self.beta[i])
                ucb = ucb_x + score_z + cb
                # ucb = ucb_x + (1-user_avg_popularity)*score_z + cb

                # if(self.mitigation is not None):
                #     # tmp_n_recommended = self.n_recommended.copy()
                #     # tmp_n_recommended[np.where(tmp_n_recommended == 0)] = 1
                #     # ucb = (1-pow(self.n_recommended/(t+1),2)) * ucb
                #     ucb = (1 - (self.n_recommended / (t + 1))) * ucb

                winner = np.argmax(ucb)
                while winner in ranking:
                    ucb[winner] = -np.inf
                    winner = np.argmax(ucb)

                ranking.append(winner)
                # ranking_set.add(winner)
                batch_features.append(z_t[winner])

                coverage = self.ranking_coverage(self.X_topical[user][ranking])

            rankings[i] = np.asarray(ranking)
            self.batch_features[i][:, :self.topic_dim] = np.asarray(batch_features)
            self.batch_features[i][:, self.topic_dim:] = self.X_latent[user][rankings[i]]
        return rankings

    # def score(self, delta):
    #     """
    #     return score for an item
    #     """
    #     return np.dot(delta[:, :self.n_z], self.beta) + np.dot(delta[:, self.n_z:], self.theta)

    # def ucb(self, delta):
    #     """
    #     return the upper confident bound of each item. This is for debugging.
    #     :param x:
    #     :return:
    #     """
    #     score = self.score(delta)
    #     z = delta[:, :self.n_z]  # topic
    #     x = delta[:, self.n_z:]  # feature
    #
    #     ZAZ = np.multiply(np.dot(z, self.A_z_inv), z).sum(axis=1)  # Z^T A_Z^-1 Z^T
    #     XAX = np.multiply(np.dot(x, self.A_x_inv), x).sum(axis=1)  # x^T A_X^-1 X^T
    #     BA_X = np.matmul(self.B.T, self.A_x_inv)  # B.T  A_x^-1
    #     ABA = np.matmul(self.A_z_inv, BA_X)  # A_z^-1 B.T A_X^-1
    #     ABABA = np.matmul(BA_X.T, ABA)
    #     s = ZAZ + XAX + np.multiply(np.dot(x, ABABA), x).sum(axis=1) - 2 * np.multiply(np.dot(z, ABA), x).sum(axis=1)
    #     cb = self.alpha * np.sqrt(s)
    #     return score + cb



import numpy as np
from .abstract_ranker import AbstractRanker
from itertools import chain, combinations
from fair_dynamic_rec.core.util.online_logistic_regression import OnlineLogisticRegression
from scipy import special
from scipy.special import binom, expit
# from scipy.optimize import minimize
from tqdm import tqdm
from joblib import Parallel, delayed

# Linear Thompson Sampling strategy for fully personalized contextual bandits, as in [Chapelle and Li, 2011]
class EARSLinearTS(AbstractRanker):
    def __init__(self, config, dataObj, parameters=None):#, user_features, n_playlists, bias=0.0, cascade_model=True, l2_reg=1, shuffle_K=0, epsilon=.0):
        super(EARSLinearTS, self).__init__(config, dataObj)

        l2_reg = float(parameters["l2_reg"]["value"])
        bias = float(parameters["bias"]["value"])

        # self.user_features = user_features
        n_dim = self.dataObj.feature_data['train_user_latent_features'].shape[1]#user_features.shape[1]
        self.n_playlists = dataObj.n_items#n_playlists
        self.models = [OnlineLogisticRegression(l2_reg, 1, n_dim, bias, 15) for i in range(dataObj.n_items)]
        self.m = np.zeros((dataObj.n_items, n_dim))
        self.m[:, -1] = bias
        self.q = np.ones((dataObj.n_items, n_dim))
        self.n_dim = n_dim
        # self.cascade_model = cascade_model
        self.shuffle_K = int(parameters["shuffle_K"]["value"])
        self.epsilon = float(parameters["epsilon"]["value"])

    def get_ranking(self, batch_users, sampled_item=None, round=None):
        user_features = np.take(self.dataObj.feature_data['train_user_latent_features'], batch_users, axis=0)
        n_users = len(batch_users)
        recos = np.zeros((n_users, self.config.list_size), dtype=np.int64)
        P_clicks = np.zeros((n_users, self.config.list_size), dtype=np.float64)
        step = 1
        u = 0
        while u < n_users:
            u_next = min(n_users, u+step)
            p_features_sampled =(np.random.normal(self.m, 1/np.sqrt(self.q), size= (u_next-u, self.n_playlists, self.n_dim)))
            step_p = p_features_sampled.dot(user_features[u:u_next].T)
            for i in range(u_next - u):
                # Extract recommendation scores for this user
                u_scores = expit(step_p[i,:,i])
                # Argsort according to negative scores to get top-K
                u_recos = np.argsort(-u_scores)[:self.config.list_size]
                recos[u+i] = u_recos
                # Extract and store scores
                P_clicks[u+i] = u_scores[u_recos]

            u += step

        # Reshuffle top-12 if specified
        if self.shuffle_K > 1:
            for u in range(n_users):
                np.random.shuffle(recos[u,:self.shuffle_K])

        # Personalised reshuffling
        if self.shuffle_K == -1:
            # For every user
            gamma = 0.9
            E_c = Parallel(n_jobs=-1)(delayed(AbstractRanker.powerset_expectation_negation_partial)(P_clicks[u], gamma, 1) for u in range(n_users))
            # print(f'Expected clicks without shuffling top-K:', np.mean(E_c), np.var(E_c))
            K = 6
            E_c = Parallel(n_jobs=-1)(delayed(AbstractRanker.powerset_expectation_negation_partial)(P_clicks[u], gamma, K) for u in range(n_users))
            # print(f'Expected clicks after shuffling Top-{K}:', np.mean(E_c), np.var(E_c))

            K_s = Parallel(n_jobs=-1)(delayed(AbstractRanker.compute_user_K_prime)(P_clicks[u], gamma, epsilon=self.epsilon) for u in range(n_users))

            for u, K in enumerate(K_s):
                np.random.shuffle(recos[u,:K])

        return recos

    def update(self, batch_users, rankings, clicks, round=None, user_round=None):#, user_ids, recos , rewards, l_init=3):
        # rewards = 2 * rewards - 1
        # batch_size = len(user_ids)
        modified_playlists = {}
        for i in range(len(batch_users)):
            user = batch_users[i]
            _clicks = self.__collect_feedback(clicks, i)
        # for i in range(batch_size):
            total_stream = len(_clicks.nonzero()[0])
            # nb_display = 0
            for p, r in zip(rankings[i], _clicks):
                # nb_display +=1
                if p not in modified_playlists:
                    modified_playlists[p] = {"X" : [], "Y" : []}
                modified_playlists[p]["X"].append(self.dataObj.feature_data['train_user_latent_features'][batch_users[i]])
                modified_playlists[p]["Y"].append(r)
                if r == 1:
                    break
        for p,v in modified_playlists.items():
            X = np.array(v["X"])
            Y = np.array(v["Y"])
            self.models[p].fit(X,Y)
            self.m[p] = self.models[p].m
            self.q[p] = self.models[p].q
        return

    def __collect_feedback(self, clicks, batch_user_id):
        """
        :param y:
        :return: the last observed position.
        """
        # With  Cascade assumption, only the first click counts.
        if self.config.feedback_model == 'cascade':
            if np.sum(clicks[batch_user_id]) == 0:
                return clicks[batch_user_id]#, self.batch_features[batch_user_id]
            first_click = np.where(clicks[batch_user_id])[0][0]
            return clicks[batch_user_id][:first_click + 1]#, self.batch_features[batch_user_id][:first_click + 1]
        elif self.config.feedback_model == 'dcm':
            if np.sum(clicks[batch_user_id]) == 0:
                return clicks[batch_user_id]#, self.batch_features[batch_user_id]
            last_click = np.where(clicks[batch_user_id])[0][-1]
            return clicks[batch_user_id][:last_click + 1]#, self.batch_features[batch_user_id][:last_click + 1]
        # all items are observed
        else:
            return clicks[batch_user_id]#, self.batch_features[batch_user_id]

# def powerset_expectation_negation_partial(R, gamma, K):
#     ''' Expected number of clicks over all possible permutations of R by counting powersets. '''
#
#     N = len(R)
#     R_neg = (1 - R).tolist()
#
#     result = 0
#     for subset in powerset(R_neg[:K], K):
#         k = len(subset)
#         prod = np.multiply.reduce(subset)
#         discount = gamma ** k
#         if k != N:
#             discount *= (1 - gamma)
#         result += prod / special.binom(K, k) * discount
#
#     for i in range(K, N):
#         discount = gamma ** (i + 1)
#         if i != (N - 1):
#             discount *= (1 - gamma)
#         result += np.multiply.reduce(R_neg[:i + 1]) * discount
#
#     return 1 - result
# ## Utilities
# def powerset(iterable, maxsize):
#     ''' From https://stackoverflow.com/questions/1482308/how-to-get-all-subsets-of-a-set-powerset '''
#     s = list(iterable)
#     return chain.from_iterable(combinations(s, r) for r in range(maxsize+1))
# def compute_user_K_prime(R, gamma = 0.9, max_K = 12, epsilon = 0.02):
#     ''' Compute the value of K' for a user, up to which we can shuffle with max epsilon loss in clicks '''
#     E_c = np.array([powerset_expectation_negation_partial(R, gamma, K) for K in range(1, max_K + 1)])
#     E_c /= E_c[0]
#     return np.searchsorted(-E_c, -(1.0 - epsilon))


from functools import reduce
from fair_dynamic_rec.core.simulators.click_simulator.cascade_model import *
from fair_dynamic_rec.core.rankers.abstract_ranker import AbstractRanker
from fair_dynamic_rec.core.util.utils import get_param_config_name
from fair_dynamic_rec.core.util.outputs import save_rankings_online_simulator, save_rewards
from tqdm import tqdm
from datetime import datetime
from scipy.special import expit
import sys
import multiprocessing as mp
from multiprocessing.pool import ThreadPool
import itertools
from sklearn.metrics.pairwise import cosine_similarity
from scipy import sparse
from sklearn.preprocessing import normalize

ClickModel = {'cascade': CascadeModel}

# This class uses user and playlist features datasets to simulate users responses to a list of recommendations

class OnlineSimulator():
    def __init__(self, config, dataObj):
        self.prng = np.random.RandomState(seed=config.seed)
        # self.data = dataObj.load_features(config)
        # config.get_optimal_ranker()
        # config.get_online_simulator_settings()
        # print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Start get optimal ranking.")
        # sys.stdout.flush()
        self.optimal_rankings, self.optimal_scores = self.get_optimal_ranking(config, dataObj)
        # print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Optimal ranking done.")
        # sys.stdout.flush()
        # ClickModel = config.feedback_model
        self.click_simulator = ClickModel[config.feedback_model](config)
        # print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Start get optimal click.")
        # sys.stdout.flush()
        self.optimal_clicks, self.optimal_rewards = self.click_simulator.get_feedback(self.optimal_scores)
        # print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Optimal click done.")
        # sys.stdout.flush()
        self.user_round = np.zeros(dataObj.n_users)
        self.overall_rewards = None

    def run(self, config, dataObj, rankers):
        print(str(datetime.now().strftime('%Y-%m-%d %H:%M:%S')) + ": Simulation started with " + str(config.rounds) + " rounds")
        sys.stdout.flush()
        overall_optimal_reward = np.zeros(config.rounds)
        self.rankers = rankers#self.set_rankers(config, dataObj)
        self.overall_rewards = np.zeros((len(self.rankers), config.rounds))
        # regret = np.zeros((len(rankers), config.rounds))
        # overall_rankings = np.zeros((len(rankers), config.rounds))
        # overall_clicks = np.zeros((len(rankers), config.rounds))
        # pool = mp.Pool(config.processor_count)
        for i in tqdm(range(0,config.rounds,config.processor_count)):
        # for i in tqdm(range(config.rounds)):
            pool = ThreadPool(processes=config.processor_count)
            # Select batch of n_users_per_round users
            user_ids = [np.random.choice(range(dataObj.n_users), config.n_users_per_round, replace=False) for j in range(config.processor_count)]
            # Select batch of items, used for factorucb algorithm
            item_ids = [np.random.choice(range(dataObj.n_items), dataObj.n_users, replace=False) for j in range(config.processor_count)]
            # user_ids = np.random.choice(range(dataObj.n_users), config.n_users_per_round, replace=False)
            for j in range(config.processor_count):
                overall_optimal_reward[i+j] = np.take(self.optimal_rewards, user_ids[j]).sum()
                self.user_round[user_ids[j]] += 1
            # overall_optimal_reward[i] = np.take(self.optimal_rewards, user_ids).sum()
            # self.user_round[user_ids] += 1
            # Iterate over all policies
            # with mp.Pool(config.processor_count) as pool:
            args = [(user_ids[j], item_ids[j], i+j, config, dataObj) for j in range(len(user_ids))]
            results = pool.map(self.run_rankers, args)
            # for j in range(len(self.rankers)):
            #     # Compute n_recos recommendations
            #     rankings = self.rankers[j]["ranker"].get_ranking(user_ids, i)
            #     # Compute rewards
            #     clicks, rewards = self.compute_rewards(config, dataObj, user_ids, rankings)
            #     # Update policy based on rewards
            #     self.rankers[j]["ranker"].update(user_ids, rankings, clicks, i, self.user_round)
            #     overall_rewards[j, i] = rewards.sum()
            #     # regret[j, i] = overall_optimal_reward.sum() - rewards.sum()
            #     # overall_rankings[j, i] = rankings
            #     # overall_clicks[j, i] = clicks
            #     save_rankings_online_simulator(config, dataObj, get_param_config_name(self.rankers[j]["config"]), i, rankings, clicks, user_ids)
            pool.close()
        save_rewards(config, self.rankers, overall_optimal_reward, self.overall_rewards)

    def run_rankers(self, args):# user_ids, i, config, dataObj
        for j in range(len(self.rankers)):
            # Compute n_recos recommendations
            rankings = self.rankers[j]["ranker"].get_ranking(args[0], args[1], args[2])
            # Compute rewards
            clicks, rewards = self.compute_rewards(args[3], args[4], args[0], rankings)
            # Update policy based on rewards
            self.rankers[j]["ranker"].update(args[0], rankings, clicks, args[2], self.user_round)
            self.overall_rewards[j, args[2]] = rewards.sum()
            # regret[j, i] = overall_optimal_reward.sum() - rewards.sum()
            # overall_rankings[j, i] = rankings
            # overall_clicks[j, i] = clicks
            save_rankings_online_simulator(args[3], args[4], get_param_config_name(self.rankers[j]["config"]), args[2], rankings, clicks, args[0])
        return

    def get_optimal_ranking(self, config, dataObj):
        self.user_features, self.item_features = AbstractRanker.get_features_for_optimal_ranker(config, dataObj)
        if config.optimal_ranker == 'naive_relevance':
            # batch_user_features = np.take(self.user_features, batch_user_ids, axis=0)
            probs = self.user_features.dot(self.item_features.T)
            optimal_rankings = np.argsort(-probs)[:, :config.list_size]
            batch_item_features = np.take(self.item_features, optimal_rankings, axis=0)
            n_users = self.user_features.shape[0]
            optimal_scores = np.zeros((n_users, config.list_size))
            # th_reward = np.zeros(n_users)
            for i in range(n_users):
                # optimal_scores[i] = 1 - reduce(lambda x, y: x * y, 1 - expit(self.user_features[i].dot(batch_item_features[i].T)))
                optimal_scores[i] = self.user_features[i].dot(batch_item_features[i].T)
            return optimal_rankings, optimal_scores
        elif config.optimal_ranker == 'topical_coverage':
            n_users = self.user_features.shape[0]
            optimal_rankings, optimal_scores = np.zeros((n_users, config.list_size), dtype=int), np.zeros((n_users, config.list_size))
            tie_breaker = self.prng.rand(len(self.item_features))
            for i in range(n_users):
                # gain_in_topic_coverage_t = []
                coverage = np.zeros(self.item_features.shape[1])
                ranking = []
                # ranking_set = set()
                for j in range(config.list_size):
                    # Line 8 - 11 of Nips 11
                    gain_in_topic_coverage = AbstractRanker.conditional_coverage(x=self.item_features, coverage=coverage)
                    probs = np.dot(gain_in_topic_coverage, self.user_features[i])
                    tmp_rank = np.lexsort((tie_breaker, -probs))
                    for tr in tmp_rank:
                        if tr not in ranking:
                            ranking.append(tr)
                            optimal_scores[i, j] = self.user_features[i].dot(gain_in_topic_coverage[tr].T)
                            # optimal_scores[i, j] = gain_in_topic_coverage[tr]
                            break
                    coverage = AbstractRanker.ranking_coverage(self.item_features[ranking])
                optimal_rankings[i] = np.asarray(ranking)
            return optimal_rankings, optimal_scores
        elif config.optimal_ranker == 'hybrid' or config.optimal_ranker == 'ears_hybrid':
            n_users = self.user_features.shape[0]
            optimal_rankings, optimal_scores = np.zeros((n_users, config.list_size), dtype=int), np.zeros((n_users, config.list_size))
            for i in range(n_users):
                # x = dataObj.feature_data['test_item_topical_features']
                # z = dataObj.feature_data['test_item_latent_features']
                # theta_star = dataObj.feature_data['test_user_topical_features']
                # mu = dataObj.feature_data['test_user_latent_features']
                # k = self.k
                # score_rel = config.lmbda * np.dot(dataObj.feature_data['test_user_latent_features'][i], dataObj.feature_data['test_item_latent_features'].T)
                score_rel = config.lmbda * np.dot(dataObj.feature_data['test_user_latent_features'][i], dataObj.feature_data['train_item_latent_features'].T)

                # delta_t = []
                gain_in_topic_coverage_t = []
                coverage = np.zeros(dataObj.feature_data['test_item_topical_features'].shape[1])
                ranking = []
                # ranking_set = set()
                for j in range(config.list_size):
                    tie_breaker = self.prng.rand(len(dataObj.feature_data['test_item_topical_features']))
                    # Line 8 - 11 of Nips 11
                    gain_in_topic_coverage = AbstractRanker.conditional_coverage(x=dataObj.feature_data['test_item_topical_features'], coverage=coverage)
                    score = score_rel + (1 - config.lmbda) * np.dot(gain_in_topic_coverage, dataObj.feature_data['test_user_topical_features'][i])
                    tmp_rank = np.lexsort((tie_breaker, -score))
                    for tr in tmp_rank:
                        if tr not in ranking:
                            ranking.append(tr)
                            # ranking_set.add(tr)
                            gain_in_topic_coverage_t.append(gain_in_topic_coverage[tr])
                            # optimal_scores[i, j] = gain_in_topic_coverage[tr]
                            score_rel[tr] = -10000
                            break
                    coverage = AbstractRanker.ranking_coverage(dataObj.feature_data['test_item_topical_features'][ranking])
                optimal_rankings[i] = np.asarray(ranking)
                gain_in_topic_coverage_t = np.asarray(gain_in_topic_coverage_t)
                delta = np.zeros((config.list_size, dataObj.feature_data['test_item_topical_features'].shape[1] + dataObj.feature_data['train_item_latent_features'].shape[1]))
                delta[:, :dataObj.feature_data['test_item_topical_features'].shape[1]] = gain_in_topic_coverage_t
                # self.optimal_scores[:, dataObj.feature_data['test_item_topical_features'].shape[1]:] = dataObj.feature_data['test_item_latent_features'][ranking]
                delta[:, dataObj.feature_data['test_item_topical_features'].shape[1]:] = dataObj.feature_data['train_item_latent_features'][ranking]
                optimal_scores[i, :] = self.user_features[i].dot(delta.T)
            return optimal_rankings, optimal_scores
        elif config.optimal_ranker == 'item_similarity':
            self.item_sim_matrix = cosine_similarity(dataObj.test_data.T)
            np.fill_diagonal(self.item_sim_matrix, 0)
            probs = dataObj.test_data.dot(self.item_sim_matrix)
            # probs = (probs-np.min(probs))/(np.max(probs)-np.min(probs))
            # self.similarity_prob = probs/np.sqrt(np.sum(probs**2))
            # self.similarity_probs = normalize(probs, axis=1, norm='l1')
            row_max = np.tile(np.amax(probs, axis=1).reshape(dataObj.test_data.shape[0],1), (1,dataObj.test_data.shape[1]))
            row_min = np.tile(np.amin(probs, axis=1).reshape(dataObj.test_data.shape[0],1), (1,dataObj.test_data.shape[1]))
            self.similarity_probs = (probs - row_min) / (row_max - row_min)
            optimal_rankings = np.argsort(-self.similarity_probs)[:, :config.list_size]
            # batch_item_features = np.take(self.item_sim_matrix, optimal_rankings, axis=0)
            n_users = dataObj.n_users #self.user_features.shape[0]
            optimal_scores = np.zeros((n_users, config.list_size))
            # th_reward = np.zeros(n_users)
            for i in range(n_users):
                # optimal_scores[i] = dataObj.test_data[i].dot(batch_item_features[i].T)
                optimal_scores[i] = self.similarity_probs[i,optimal_rankings[i]]
            return optimal_rankings, optimal_scores

    def compute_rewards(self, config, dataObj, batch_users, rankings):
        batch_user_features = np.take(self.user_features, batch_users, axis=0)
        batch_item_features = np.take(self.item_features, rankings, axis=0)
        if config.optimal_ranker == 'naive_relevance':
            n_users = batch_user_features.shape[0]
            scores = np.zeros((n_users, config.list_size))
            # reward = np.zeros(n_users)
            for i in range(n_users):
                scores[i] = batch_user_features[i].dot(batch_item_features[i].T)
                # reward[i] = 1 - reduce(lambda x, y: x * y, 1 - probas)
            return self.click_simulator.get_feedback(scores)
        elif config.optimal_ranker == 'topical_coverage':
            n_users = batch_user_features.shape[0]
            scores = np.zeros((n_users, config.list_size))
            for i in range(n_users):
                topic_coverage = self.__convert_to_topic_coverage(batch_item_features[i])
                scores[i] = batch_user_features[i].dot(topic_coverage.T)
            return self.click_simulator.get_feedback(scores)
        elif config.optimal_ranker == 'hybrid':
            item_topic = self.item_features[:, :self.item_features.shape[1]-config.get_latent_feature_dim()]
            item_latent = self.item_features[:, self.item_features.shape[1] - config.get_latent_feature_dim():]
            n_users = batch_user_features.shape[0]
            scores = np.zeros((n_users, config.list_size))
            for i in range(n_users):
                # topic_coverage = self.__convert_to_topic_coverage(dataObj.feature_data['test_item_topical_features'][rankings[i]])
                topic_coverage = self.__convert_to_topic_coverage(item_topic[rankings[i]])
                # item_full_feature = np.zeros((config.list_size, dataObj.feature_data['test_item_topical_features'].shape[1] + dataObj.feature_data['train_item_latent_features'].shape[1]))
                item_full_feature = np.zeros((config.list_size, item_topic.shape[1] + item_latent.shape[1]))
                # item_full_feature[:, :dataObj.feature_data['test_item_topical_features'].shape[1]] = topic_coverage
                item_full_feature[:, :item_topic.shape[1]] = topic_coverage
                # item_full_feature[:, dataObj.feature_data['test_item_topical_features'].shape[1]:] = dataObj.feature_data['train_item_latent_features'][rankings[i]]
                item_full_feature[:, item_topic.shape[1]:] = item_latent[rankings[i]]
                scores[i] = batch_user_features[i].dot(item_full_feature.T)
            return self.click_simulator.get_feedback(scores)
        elif config.optimal_ranker == 'item_similarity':
            n_users = len(batch_users)
            scores = np.zeros((n_users, config.list_size))
            for i in range(n_users):
                scores[i] = self.similarity_probs[batch_users[i],rankings[i]]
            return self.click_simulator.get_feedback(scores)
        # elif config.optimal_ranker == 'ears_hybrid':
        #     n_users = batch_user_features.shape[0]
        #     scores = np.zeros((n_users, config.list_size))
        #     for i in range(n_users):
        #         topic_coverage = self.__convert_to_topic_coverage(dataObj.feature_data['test_item_topical_features'][rankings[i]])
        #         item_full_feature = np.zeros((config.list_size, dataObj.feature_data['test_item_topical_features'].shape[1] + dataObj.feature_data['train_item_latent_features'].shape[1]))
        #         item_full_feature[:, :dataObj.feature_data['test_item_topical_features'].shape[1]] = topic_coverage
        #         item_full_feature[:, dataObj.feature_data['test_item_topical_features'].shape[1]:] = dataObj.feature_data['test_item_latent_features'][rankings[i]]
        #         scores[i] = batch_user_features[i].dot(item_full_feature.T)
        #     return self.click_simulator.get_feedback(scores)

    def __convert_to_topic_coverage(self, x):
        k, d = x.shape
        score = []
        coverage = np.zeros(d)
        for idx, topic in enumerate(x):
            score.append(AbstractRanker.conditional_coverage(x=topic, coverage=coverage))
            coverage = AbstractRanker.ranking_coverage(x[:idx+1])
        return np.asarray(score)






    # Computes expected reward for each user given their recommendations
    def compute_theoretical_rewards(self, batch_user_ids, batch_recos):
        batch_user_features = np.take(self.user_features, batch_user_ids, axis=0)
        batch_playlist_features = np.take(self.playlist_features, batch_recos, axis=0)
        n_users = len(batch_user_ids)
        th_reward = np.zeros(n_users)
        for i in range(n_users):
            probas = expit(batch_user_features[i].dot(batch_playlist_features[i].T))
            th_reward[i] = 1 - reduce(lambda x, y: x * y, 1 - probas)
        return th_reward

    # Computes list of n recommendations with highest expected reward for each user
    def compute_optimal_recos(self, batch_user_ids, n):
        batch_user_features = np.take(self.user_features, batch_user_ids, axis=0)
        probas = batch_user_features.dot(self.playlist_features.T)
        optim = np.argsort(-probas)[:, :n]
        return optim

    # Computes highest expected reward for each user
    def compute_optimal_theoretical_rewards(self):
        n_users = self.user_features.shape[0]
        u = 0
        step = 100000
        while u < n_users:
            users_ids = range(u, min(n_users, u + step))
            opt_recos = self.compute_optimal_recos(users_ids, self.n_recos)
            opt_rewards = self.compute_theoretical_rewards(users_ids, opt_recos)
            self.th_rewards[u:min(n_users, u + step)] = opt_rewards
            u += step
        return

        # Computes list of n recommendations with highest expected reward for each segment

    def compute_segment_optimal_recos(self, n):
        n_segments = len(np.unique(self.user_segment))
        segment_recos = np.zeros((n_segments, n), dtype=np.int64)
        for i in range(n_segments):
            mean_probas = np.mean(expit(
                np.take(self.user_features, np.where(self.user_segment == i)[0], axis=0).dot(self.playlist_features.T)),
                                  axis=0)
            reward = 1 - reduce(lambda x, y: x * y, 1 + np.sort(-mean_probas)[:n])
            segment_recos[i] = np.argsort(-mean_probas)[:n]
        return segment_recos

    # Computes highest expected reward for each segment
    def compute_segment_optimal_theoretical_rewards(self):
        n_users = self.user_features.shape[0]
        u = 0
        step = 100000
        segment_recos = self.compute_segment_optimal_recos(self.n_recos)
        while u < n_users:
            users_ids = range(u, min(n_users, u + step))
            user_segment = np.take(self.user_segment, users_ids)
            opt_recos = np.take(segment_recos, user_segment, axis=0)
            opt_rewards = self.compute_theoretical_rewards(users_ids, opt_recos)
            self.th_segment_rewards[u:min(n_users, u + step)] = opt_rewards
            u += step
        return

        # Given a list of users and their respective list of recos (each of size self.n_recos), computes

    # corresponding simulated reward
    def simulate_batch_users_reward(self, batch_user_ids, batch_recos):

        # First, compute probability of streaming each reco and draw rewards accordingly
        batch_user_features = np.take(self.user_features, batch_user_ids, axis=0)
        batch_playlist_features = np.take(self.playlist_features, batch_recos, axis=0)
        n_users = len(batch_user_ids)
        n = len(batch_recos[0])
        probas = np.zeros((n_users, n))
        for i in range(n_users):
            probas[i] = expit(
                batch_user_features[i].dot(batch_playlist_features[i].T))  # probability to stream each reco
        rewards = np.zeros((n_users, n))
        i = 0
        rewards_uncascaded = np.random.binomial(1, probas)  # drawing rewards from probabilities
        positive_rewards = set()

        # Then, for each user, positive rewards after the first one are set to 0 (and playlists as "unseen" subsequently)
        # to imitate a cascading browsing behavior
        # (nonetheless, users can be drawn several times in the batch of a same round ; therefore, each user
        # can have several positive rewards - i.e. stream several playlists - in a same round, consistently with
        # the multiple-plays framework from the paper)
        nz = rewards_uncascaded.nonzero()
        for i in range(len(nz[0])):
            if nz[0][i] not in positive_rewards:
                rewards[nz[0][i]][nz[1][i]] = 1
                positive_rewards.add(nz[0][i])
        return rewards




# CLICKMODEL_MAP = {'CM': CascadeSimulator,
#                   'cascademodel': CascadeSimulator,
#                   'cm': CascadeSimulator,
#                   'PBM': PBMSimulator,
#                   'pbm': PBMSimulator,
#                   'positionbasemodel': PBMSimulator,
#                   'DCM': DCMSimulator,
#                   'dcm': DCMSimulator,
#                   'DBM': DBMSimulator,
#                   'dbm': DBMSimulator,
#                   'documentbasemodel': DBMSimulator
#                   }
#
#
# class OnlineSimulator1(object):
#
#     name ='OnlineSimulator'
#
#     def __init__(self, dataset, sim_args, q_id):
#         """
#
#         :param dataset: dict contains train_x, test_x, theta_star
#         :param ClickModel: class: click mode
#         :param sim_args: the simulation arguments from utils/my_parser.py
#                         we use k: n_position, iteration, seed: random seed, output: output dict
#         """
#         self.sim_args = sim_args
#         self.iteration = sim_args.iteration
#         self.k = sim_args.K
#         self.prng = np.random.RandomState(seed=sim_args.seed)
#         # this one is the input for the online algorithms
#         self.train_x = dataset['train_x']
#         # the following two define the click model
#         self.test_x = dataset['test_x']
#         self.theta_star = dataset['theta_star']
#
#         self.n_topics = self.train_x.shape[1]
#         self.n_items = self.train_x.shape[0]
#         ClickModel = CLICKMODEL_MAP[self.sim_args.ClickModel]
#         self.click_simulator = ClickModel(theta_star=self.theta_star, seed=sim_args.seed)
#         # find the best ranking based on the testing feature space.
#         self.best_ranking, self.best_delta = self.__greedy_search(self.test_x, self.k)
#
#         self.q_id = q_id
#
#     def __greedy_search(self, x, k):
#         """
#         search for the optimal ranking
#         :param x: input feature
#         :param k: number of postions
#         :return: optimal ranking
#         """
#         delta_t = []
#         coverage = np.zeros(self.n_topics)
#         ranking = []
#         ranking_set = set()
#         for i in range(k):
#             tie_breaker = self.prng.rand(len(x))
#             # Line 8 - 11 of Nips 11
#             delta = AbstractRanker.conditional_coverage(x=x, coverage=coverage)
#             score = np.dot(delta, self.theta_star)
#             tmp_rank = np.lexsort((tie_breaker, -score))
#             for tr in tmp_rank:
#                 if tr not in ranking_set:
#                     ranking.append(tr)
#                     ranking_set.add(tr)
#                     delta_t.append(delta[tr])
#                     break
#             coverage = AbstractRanker.ranking_coverage(x[ranking])
#         return ranking, np.asarray(delta_t)
#
#     def __convert_to_topic_coverage(self, x):
#         k, d = x.shape
#         delta_t = []
#         coverage = np.zeros(d)
#         for idx, topic in enumerate(x):
#             delta = AbstractRanker.conditional_coverage(x=topic, coverage=coverage)
#             delta_t.append((delta))
#             coverage = AbstractRanker.ranking_coverage(x[:idx+1])
#         return np.asarray(delta_t)
#
#     def run(self, rankers, save_results=False):
#         if type(rankers) is not list:
#             rankers = [rankers]
#
#         regret = {}
#         reward = {}
#         for ranker in rankers:
#             regret[ranker.name] = np.zeros(self.iteration)
#             reward[ranker.name] = np.zeros(self.iteration)
#
#         for i in range(self.iteration):
#             if self.sim_args.same_coins:
#                 self.click_simulator.set_coins(self.k)
#
#             best_clicks, best_reward = self.click_simulator.get_feedback(self.best_delta)
#             for ranker in rankers:
#                 ranking, delta = ranker.get_ranking(self.train_x, self.k, i)
#                 ranker.n_recommended[np.array(ranking)] = ranker.n_recommended[np.array(ranking)] + 1
#                 delta = self.__convert_to_topic_coverage(self.test_x[ranking])
#                 """
#                 The click is from the click simulator. So it is defined by the testing part. w_test and theta_test.
#                 Bug from the click log, the bias from train_x to test_x is large. Here, I still use train.
#                 """
#                 clicks, t_reward = self.click_simulator.get_feedback(delta)
#
#                 if (sum(clicks) > 0):
#                     save_history(i, self.q_id, ranking, np.array(ranking)[clicks])
#                 else:
#                     save_history(i, self.q_id, ranking, [])
#
#                 ranker.update(y=clicks)
#                 reward[ranker.name][i] = t_reward
#                 regret[ranker.name][i] = best_reward - t_reward
#
#             if self.sim_args.same_coins:
#                 self.click_simulator.del_coins()
#         # if save_results:
#         #     self.__save_results(rankers=rankers, reward=reward, regret=regret)
#
#         return reward, regret
#
#     def save_results(self, q_id, rankers, reward, regret):
#         """
#         save results to the self.sim_args.output director
#         the name is ranker.name+ranker parameters + random seed + save date
#         :param rankers: same as self.run()
#         :param reward: output of self.run()
#         :param regret: output of self.run()
#         :return: save results to json file.
#         """
#         # Saving directory
#         if self.sim_args.output[-1] == '/':
#             prefix = self.sim_args.output + \
#                      '/'.join([self.sim_args.data_name, self.sim_args.ClickModel, 'norm-'+str(self.sim_args.normalized),
#                                'rep'+str(self.sim_args.iteration),
#                                'pos'+str(self.sim_args.K), 'topic'+str(self.sim_args.n_topic)]) + '/' + str(q_id) + '/'
#         else:
#             prefix = self.sim_args.output + '/' + \
#                      '/'.join([self.sim_args.data_name, self.sim_args.ClickModel, 'norm-'+str(self.sim_args.normalized),
#                                'rep'+str(self.sim_args.iteration),
#                                'pos'+str(self.sim_args.K), 'topic'+str(self.sim_args.n_topic)]) + '/' + str(q_id) + '/'
#
#         if not os.path.exists(prefix):
#             os.makedirs(prefix)
#
#         suffix = 'seed-' + str(self.sim_args.seed) + \
#                  '-' + str(datetime.datetime.now().date()) + \
#                  '-' + str(datetime.datetime.now().time())[:8].replace(':', '-') \
#                  + '.js'
#
#         for ranker in rankers:
#             save_name = prefix + ranker.name + '-alpha%.2f-sigma%.2f-' % (ranker.alpha, ranker.sigma) + suffix
#             objs = {'reward': reward[ranker.name].tolist(),
#                     'regret': regret[ranker.name].tolist()
#                     }
#             with open(save_name, 'w') as f:
#                 json.dump(objs, f)
#
# class OnlineNaiveSimulator(object):
#
#     name ='OnlineNaiveSimulator'
#
#     def __init__(self, dataset, sim_args, q_id):
#         """
#
#         :param dataset: dict contains train_x, test_x, theta_star
#         :param ClickModel: class: click mode
#         :param sim_args: the simulation arguments from utils/my_parser.py
#                         we use k: n_position, iteration, seed: random seed, output: output dict
#         """
#         self.sim_args = sim_args
#         self.iteration = sim_args.iteration
#         self.k = sim_args.K
#         self.prng = np.random.RandomState(seed=sim_args.seed)
#         # this one is the input for the online algorithms
#         self.train_x = dataset['train_x']
#         # the following two define the click model
#         self.test_x = dataset['test_x']
#         self.theta_star = dataset['theta_star']
#
#         self.n_topics = self.train_x.shape[1]
#         self.n_items = self.train_x.shape[0]
#         ClickModel = CLICKMODEL_MAP[self.sim_args.ClickModel]
#         self.click_simulator = ClickModel(theta_star=self.theta_star, seed=sim_args.seed)
#         # find the best ranking based on the testing feature space.
#         self.best_ranking, self.best_delta = self.__greedy_search(self.test_x, self.k)
#
#         self.q_id = q_id
#
#     def __greedy_search(self, x, k):
#         """
#         search for the optimal ranking
#         :param x: input feature
#         :param k: number of postions
#         :return: optimal ranking
#         """
#         delta_t = []
#         coverage = np.zeros(self.n_topics)
#         ranking = []
#         ranking_set = set()
#         for i in range(k):
#             tie_breaker = self.prng.rand(len(x))
#             # Line 8 - 11 of Nips 11
#             delta = AbstractRanker.conditional_coverage(x=x, coverage=coverage)
#             score = np.dot(delta, self.theta_star)
#             tmp_rank = np.lexsort((tie_breaker, -score))
#             for tr in tmp_rank:
#                 if tr not in ranking_set:
#                     ranking.append(tr)
#                     ranking_set.add(tr)
#                     delta_t.append(delta[tr])
#                     break
#             coverage = AbstractRanker.ranking_coverage(x[ranking])
#         return ranking, np.asarray(delta_t)
#
#     def __convert_to_topic_coverage(self, x):
#         k, d = x.shape
#         delta_t = []
#         coverage = np.zeros(d)
#         for idx, topic in enumerate(x):
#             delta = AbstractRanker.conditional_coverage(x=topic, coverage=coverage)
#             delta_t.append((delta))
#             coverage = AbstractRanker.ranking_coverage(x[:idx+1])
#         return np.asarray(delta_t)
#
#     def run(self, rankers, save_results=False):
#         if type(rankers) is not list:
#             rankers = [rankers]
#
#         regret = {}
#         reward = {}
#         for ranker in rankers:
#             regret[ranker.name] = np.zeros(self.iteration)
#             reward[ranker.name] = np.zeros(self.iteration)
#
#         for i in range(self.iteration):
#             if self.sim_args.same_coins:
#                 self.click_simulator.set_coins(self.k)
#
#             best_clicks, best_reward = self.click_simulator.get_feedback(self.best_delta)
#             for ranker in rankers:
#                 ranking, delta = ranker.get_ranking(self.train_x, self.k, i)
#                 ranker.n_recommended[np.array(ranking)] = ranker.n_recommended[np.array(ranking)] + 1
#                 delta = self.__convert_to_topic_coverage(self.test_x[ranking])
#                 """
#                 The click is from the click simulator. So it is defined by the testing part. w_test and theta_test.
#                 Bug from the click log, the bias from train_x to test_x is large. Here, I still use train.
#                 """
#                 clicks, t_reward = self.click_simulator.get_feedback(delta)
#
#                 if (sum(clicks) > 0):
#                     save_history(i, self.q_id, ranking, np.array(ranking)[clicks])
#                 else:
#                     save_history(i, self.q_id, ranking, [])
#
#                 ranker.update(y=clicks)
#                 reward[ranker.name][i] = t_reward
#                 regret[ranker.name][i] = best_reward - t_reward
#
#             if self.sim_args.same_coins:
#                 self.click_simulator.del_coins()
#         # if save_results:
#         #     self.__save_results(rankers=rankers, reward=reward, regret=regret)
#
#         return reward, regret
#
#     def save_results(self, q_id, rankers, reward, regret):
#         """
#         save results to the self.sim_args.output director
#         the name is ranker.name+ranker parameters + random seed + save date
#         :param rankers: same as self.run()
#         :param reward: output of self.run()
#         :param regret: output of self.run()
#         :return: save results to json file.
#         """
#         # Saving directory
#         if self.sim_args.output[-1] == '/':
#             prefix = self.sim_args.output + \
#                      '/'.join([self.sim_args.data_name, self.sim_args.ClickModel, 'norm-'+str(self.sim_args.normalized),
#                                'rep'+str(self.sim_args.iteration),
#                                'pos'+str(self.sim_args.K), 'topic'+str(self.sim_args.n_topic)]) + '/' + str(q_id) + '/'
#         else:
#             prefix = self.sim_args.output + '/' + \
#                      '/'.join([self.sim_args.data_name, self.sim_args.ClickModel, 'norm-'+str(self.sim_args.normalized),
#                                'rep'+str(self.sim_args.iteration),
#                                'pos'+str(self.sim_args.K), 'topic'+str(self.sim_args.n_topic)]) + '/' + str(q_id) + '/'
#
#         if not os.path.exists(prefix):
#             os.makedirs(prefix)
#
#         suffix = 'seed-' + str(self.sim_args.seed) + \
#                  '-' + str(datetime.datetime.now().date()) + \
#                  '-' + str(datetime.datetime.now().time())[:8].replace(':', '-') \
#                  + '.js'
#
#         for ranker in rankers:
#             save_name = prefix + ranker.name + '-alpha%.2f-sigma%.2f-' % (ranker.alpha, ranker.sigma) + suffix
#             objs = {'reward': reward[ranker.name].tolist(),
#                     'regret': regret[ranker.name].tolist()
#                     }
#             with open(save_name, 'w') as f:
#                 json.dump(objs, f)
#
#
# if __name__ == '__main__':
#     args = SimulationArgumentParser()
#     sim_args = args.parse_args('-K 2 --output results/ --data synthetic --seed 2'.split())
#
#     SimulationArgumentParser.print(sim_args)
#
#     with open('../data/synthetic.pkl', 'rb') as f:
#         dataset = pk.load(f)
#
#     d = dataset['train_x'].shape[1]
#     rankers = [CascadeLSB(d=d, sigma=.1, alpha=.8, seed=sim_args.seed),
#                LSBGreedy(d=d, sigma=.1, alpha=.8, seed=sim_args.seed)
#               ]
#
#     sim = OnlineSimulator(dataset=dataset, sim_args=sim_args)
#     reward, regret = sim.run(rankers=rankers, save_results=True)
#
#     for key in reward:
#         print(key, (np.cumsum(reward[key])/np.arange(1, 1+sim_args.iteration))[-10:])
#
#     for key in regret:
#         print(key, np.cumsum(regret[key])[-10:])
#
